/* ----
 * ---- file   : reg.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 10Sep2013, 11Sep2013, 14Sep2013, 15Sep2013
 * ----
 * ----
 */

#define __NO_VERSION__

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/io.h>

#include "../include/types.h"

#include "kmod.h"
#include "reg.h"


#define REG_TRACE defined


/* (note) it seems that the kernel 3.2.45 ioremap_nocache() call sometimes returns illegal addresses 
 *         (mmap()ing a physical addr. in userland also goes haywire "randomly".
 *
 *         i.e. avoid frequent mapping/unmapping if possible, mapping an entire register set
 *          at once during init() seems to work (most of the time)
 *
 */


/*--------------------------------------------------------------------------- reg32_write() */
void reg32_write(sU32 _physAddr, sU32 _val) {
   void __iomem *virtAddr = ioremap_nocache(_physAddr, 4);

   if(NULL != virtAddr)
   {
#ifdef REG_TRACE
      if(c64_debug_level >= 20)
      {
         printk(KERN_INFO "xxx loc_reg32_write: physAddr=0x%08x virtAddr=%08x val=0x%08x\n",
                _physAddr,
                (sU32)virtAddr,
                _val
                );
      }
#endif // REG_TRACE

      *((sU32*)virtAddr) = _val;

      iounmap(virtAddr);
   }
  
}


/*--------------------------------------------------------------------------- reg32_read() */
sU32 reg32_read(sU32 _physAddr) {
   sU32 ret;
   void __iomem *virtAddr = ioremap_nocache(_physAddr, 4);

   if(NULL != virtAddr)
   {
      ret = *((sU32*)virtAddr);

#ifdef REG_TRACE
      if(c64_debug_level >= 20)
      {
         printk(KERN_INFO "xxx loc_reg32_read: physAddr=0x%08x val=0x%08x\n",
                _physAddr,
                ret
                );
      }
#endif // REG_TRACE

      iounmap(virtAddr);
   }
   else
   {
      ret = 0xCCccCCccu;
   }

   return ret;
}


/*--------------------------------------------------------------------------- reg32_bit_clear() */
void reg32_bit_clear(sU32 _physAddr, sU32 _bit) {

   void __iomem *virtAddr = ioremap_nocache(_physAddr, 4);

   if(NULL != virtAddr)
   {
      *((sU32*)virtAddr) &= ~(1 << _bit);

      iounmap(virtAddr);
   }
  
}


/*--------------------------------------------------------------------------- reg32_bit_set() */
void reg32_bit_set(sU32 _physAddr, sU32 _bit) {
   void __iomem *virtAddr = ioremap_nocache(_physAddr, 4);

   if(NULL != virtAddr)
   {
#ifdef REG_TRACE
      if(c64_debug_level >= 20)
      {
         printk(KERN_INFO "xxx loc_reg32_bit_set:   addr=0x%08x bit=%u oldVal=0x%08x\n",
                _physAddr,
                _bit,
                *((sU32*)virtAddr)
                );
      }
#endif // REG_TRACE
             
      
      *((sU32*)virtAddr) |= (1 << _bit);

#ifdef REG_TRACE
      if(c64_debug_level >= 20)
      {
         printk(KERN_INFO "xxx 2 loc_reg32_bit_set: addr=0x%08x bit=%u newVal=0x%08x\n",
                _physAddr,
                _bit,
                *((sU32*)virtAddr)
                );
      }
#endif // REG_TRACE

      iounmap(virtAddr);
   }
  
}
/* ----
 * ---- file   : mlb.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 10Sep2013, 11Sep2013, 14Sep2013, 15Sep2013
 * ----
 * ----
 */

#define HAVE_BROKEN_MAILBOX defined


#define __NO_VERSION__

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <linux/../../arch/arm/include/asm/cacheflush.h>


#include "../include/types.h"
#include "../include/msg.h"
#include "../include/dsp_common.h"
#define C64_KBUILD defined
#include "../include/dsp_config.h"

#include "../include/arch/dm3730/dm3730.h"

#include "kmod.h"
#include "reg.h"
#include "dev.h"
#include "mlb.h"

////extern void omap_writel(u32 v, u32 pa); /* (todo) use proper include file */
////extern u32  omap_readl(u32 pa); /* (todo) use proper include file */


#define MLB_MSG_POOL_SIZE  (RESET_VECTOR_MSGSIZE / sizeof(dsp_msg_t))  // => 248


/*--------------------------------------------------------------------------- module vars */
static volatile sU32 mlb_irq_count = 0;
static volatile sU32 mlb_bh_count = 0;

static volatile sU32 mlb_num_recvd = 0;
static volatile sU32 mlb_num_sent = 0;  // == num revd by the DSP

static sU8 __iomem *mlb_base_virt;  /* also see "dsp_common.h" */

#define MLB_REG(a)  (*(volatile sU32*)( ((a) - MLB_BASE) + mlb_base_virt ))

static volatile sU32 mlb_inbox[MLB_FIFO_SIZE]; /* written by the IRQ handler. actually only one message. */

////static volatile sU32 mlb_dsp_inbox; // workaround for FIFO clear fail on DSP-side

static sU8 __iomem *msg_base_virt;

#define MSG_BASE  (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE)
#define MSG_ADDR(a) (((a) - MSG_BASE) + msg_base_virt)
#define MSG_REG(a)  (*(volatile sU32 __iomem*)( MSG_ADDR(a) ))


/* (note) calling ioremap_nocache() / iounmap() in interrupt handler freezes the OS */
//#define MSG_REG_READ(a) reg32_read(a)
//#define MSG_REG_WRITE(a, b) reg32_write(a, b)

// crashes:
/////#define MSG_REG_READ(a) omap_readl(a)
/////#define MSG_REG_WRITE(a, b) omap_writel(b, a)

// works but access still is cached (DSP sometimes sees wrong data):
#define MSG_REG_READ(a) __raw_readl((const volatile void __iomem*)(MSG_ADDR(a)))
#define MSG_REG_WRITE(a, b) __raw_writel(b, (volatile void __iomem*)MSG_ADDR(a))


//#define MSG_REG_READ(a) MSG_REG(a)
//#define MSG_REG_WRITE(a, b) MSG_REG(a) = b


#ifdef PANDORA_VID2_MSG_HACK
static sU8 __iomem *hack_base_virt;
#define HACK_ADDR(a) (((a) - DSP_HACK_BASE) + hack_base_virt)
#define HACK_REG(a)  (*(volatile sU32 __iomem*)( HACK_ADDR(a) ))

#define HACK_REG_READ(a) __raw_readl((const volatile void __iomem*)(HACK_ADDR(a)))
#define HACK_REG_WRITE(a, b) __raw_writel(b, (volatile void __iomem*)HACK_ADDR(a))
#endif /* PANDORA_VID2_MSG_HACK */


////static volatile sUI msg_idx;  // current send/recv message index

static sU32 msg_serial;  /* required for buggy HW workaround so the DSP can identify unique messages */

static atomic_t at_b_ready_to_send = ATOMIC_INIT(S_FALSE);  /* false while waiting for reply from DSP */


/*--------------------------------------------------------------------------- loc_mlb_reset() */
static void loc_mlb_reset(void) {

   /* Disable GPP mailbox interrupts */
   MLB_REG(MLB_IRQENABLE_0) = 0;
   
   /* Disable DSP mailbox interrupts */
   MLB_REG(MLB_IRQENABLE_1) = 0;

   /* Mailbox software reset + set SIDLEMODE to "smart-idle" and autoidle to
    *  automatic interface clocking (based on L4-core interface activity)
    */
   MLB_REG(MLB_SYSCONFIG) = 
      (MLB_REG(MLB_SYSCONFIG) & ~((1u << 3) | (1u << 4))) |
      (2u << 3) |  // SIDLEMODE
      (1u << 0) |  // AUTOIDLE
      (1u << 1) ;  // SOFTRESET

   /* Store message serial so the DSP can distinguish between interrupt spam
    *  (on my HW, it always receives at least two interrupts, even with the workarounds)
    *  and "real" messages.
    */
   if(c64_debug_level >= 20)
   {
      printk(KERN_DEBUG "writing msg_serial=0x%08x to 0x%08x\n", msg_serial, DSP_MLB_RAM_MSGSERIAL);
   }

#ifdef PANDORA_VID2_MSG_HACK
   if(c64_debug_level >= 20)
   {
      printk(KERN_DEBUG "write msgserial hack_base_virt=0x%08x, MSGSERIAL_phys=0x%08x MSGSERIAL_virt=0x%08x\n",
             (sU32)hack_base_virt,
             DSP_MLB_RAM_MSGSERIAL,
             (sU32)HACK_ADDR(DSP_MLB_RAM_MSGSERIAL)
             );
   }
   HACK_REG_WRITE(DSP_MLB_RAM_MSGSERIAL, msg_serial);
#else
   MSG_REG_WRITE(DSP_MLB_RAM_MSGSERIAL, msg_serial);
#endif /* PANDORA_VID2_MSG_HACK */

   /* Wait for mailbox software reset */
   while(1 != MLB_REG(MLB_SYSSTATUS))
   {
      /* Wait for RESETDONE bit */
   }
}


/*--------------------------------------------------------------------------- loc_mlb_reset_2() */
static void loc_mlb_reset_2(void) {
   /* GPP-side:
    *  Setup mailbox 0 for sending messages from the GPP to the DSP, and
    *        mailbox 1 for receiving messages from the DSP
    */
#if 1
   MLB_REG(MLB_IRQENABLE_0) = 
      (
         //(1u << 1) |  // Generate IRQ when msg was delivered to the DSP and FIFO slot becomes avail.
         (1u << 2)    // set NEWMSGENABLEUUMB1 --> Generate IRQ when msg was received from the DSP
       );
#else
   MLB_REG(MLB_IRQENABLE_1) = 
      (
         //(1u << 3) |  // Generate IRQ when msg was delivered to the GPP and FIFO slot becomes avail.
         (1u << 0)    // set NEWMSGENABLEUUMB0 --> Generate IRQ when msg was received from the GPP
       );
#endif
   
   
#if 1
   /* DSP-side:
    *  Setup mailbox 0 for receiving messages from the DSP to the GPP, and
    *        mailbox 1 for sending messages to the GPP
    */
#if 1
   MLB_REG(MLB_IRQENABLE_1) = 
      (
         //(1u << 3) |  // Generate IRQ when msg was delivered to the GPP and FIFO slot becomes avail.
         (1u << 0)    // set NEWMSGENABLEUUMB0 --> Generate IRQ when msg was received from the GPP
       );
#else
   MLB_REG(MLB_IRQENABLE_0) = 
      (
         //(1u << 1) |  // Generate IRQ when msg was delivered to the GPP and FIFO slot becomes avail.
         (1u << 2)    // set NEWMSGENABLEUUMB1 --> Generate IRQ when msg was received from the GPP
       );
#endif
#endif
}


/*--------------------------------------------------------------------------- bh_handler__mlb */
static void bh_handler__mlb(unsigned long _userData) {
   /* Bottom half IRQ handler (not used ATM, have to figured out why tasklet_schedule() freezes the OS.
       Probably not that important, anyway.
    */
   (void)_userData;

   mlb_bh_count++;

   ////wake_up_interruptible(&mlb_queue);
}

static DECLARE_TASKLET(c64_mlb_tasklet, bh_handler__mlb, 0);


/*--------------------------------------------------------------------------- irq_handler__mlb */
static irqreturn_t irq_handler__mlb(int _irq,
                                    void * _devId // priv data ptr (NULL)
                                    ) {
   /* Top half IRQ handler (actually does the entire IRQ handling ATM) */

   mlb_irq_count++;

   if(c64_debug_level >= 20)
   {
      printk(KERN_INFO "irq_handler_mlb\n");
   }

   /* Read messages */
   {
      sUI numMsg = MLB_REG(MLB_MSGSTATUS_1) & 7; // mask NBOFMSGMB field
      //sUI numMsg = MLB_REG(MLB_MSGSTATUS_0) & 7; // mask NBOFMSGMB field

      if(numMsg > 0)
      {
         if(numMsg <= MLB_FIFO_SIZE)  // sanity check
         {
            sUI fifoIdx = 0;

            while(fifoIdx < numMsg)
            {
               mlb_inbox[fifoIdx] = MLB_REG(MLB_MESSAGE_1);
               //mlb_inbox[fifoIdx] = MLB_REG(MLB_MESSAGE_0);

               //if(mlb_num_recvd < 10) // don't flood the klog
               {
                  mlb_num_recvd++;

                  if(c64_debug_level >= 16)
                  {
                     /* Uhh.. printk in an IRQ handler.. ;) */ 
                     printk(KERN_DEBUG "c64: mlb_inbox[%u/%u] = 0x%08x\n",
                            (fifoIdx + 1),
                            numMsg,
                            mlb_inbox[fifoIdx]
                            );
                  }
               }
               
               // Next message
               fifoIdx++;
            }
         }
         /* else: numMsg exceeds FIFO size */
      }
      /* else: no messages to read */
   }

   /* Acknowledge interrupt */
   MLB_REG(MLB_IRQSTATUS_0) = (1u << 2);  // clear NEWMSGSTATUSUUMB1
   //MLB_REG(MLB_IRQSTATUS_0) = 15;  // clear all bits (no change, still getting interrupt spam)
   ////MLB_REG(MLB_IRQSTATUS_1) |= (1 << 0);  // clear NEWMSGSTATUSUUMB0
   
   ////tasklet_schedule(&c64_mlb_tasklet); // freezes system

   /* Disable mailbox interrupts to prevent interrupt spam
    *  (workaround for buggy hardware)
    */
   MLB_REG(MLB_IRQENABLE_0) = 0;

   /* Copy DSP reply to client structure */
   {
#if 0
      dev_handle_reply((const volatile dsp_msg_t *)MSG_ADDR(DSP_MLB_RAM_MSGTODSP), /* for client handle */
                       (const volatile dsp_msg_t *)MSG_ADDR(DSP_MLB_RAM_MSGTOGPP)
                       );
#else
      dsp_msg_t reply;

#if 0
      flush_cache_all();
#endif

#ifdef PANDORA_VID2_MSG_HACK
      /* also see "dsp_commmon.h" */
      reply.c64_msg.data[0].ret = HACK_REG_READ(DSP_MLB_RAM_MSGTOGPP_RET1);
      reply.c64_msg.data[1].ret = HACK_REG_READ(DSP_MLB_RAM_MSGTOGPP_RET2);
#else
      /* (note) calling ioremap_nocache() / iounmap() in interrupt handler freezes the OS 
       * (note) this does not work occasionally since the memory seems cached, after all.
       */
      reply.c64_msg.data[0].ret = MSG_REG(DSP_MLB_RAM_MSGTOGPP + 4u + 4u);
      reply.c64_msg.data[1].ret = MSG_REG(DSP_MLB_RAM_MSGTOGPP + 4u + 8u);

#endif /* PANDORA_VID2_MSG_HACK */

      dev_handle_reply((const volatile dsp_msg_t *)MSG_ADDR(DSP_MLB_RAM_MSGTODSP),
                       &reply
                       );

#endif
   }

   /* Ready to send the next message */
   atomic_set(&at_b_ready_to_send, S_TRUE);

   return IRQ_HANDLED;
}


/*--------------------------------------------------------------------------- mlb_send_to_dsp() */
int mlb_send_to_dsp(void *_client, const c64_msg_t *_msg) {
   int ret;

   if(S_TRUE == atomic_read(&at_b_ready_to_send) || (NULL == _client)/*force testmsg*/)
   {
      atomic_set(&at_b_ready_to_send, S_FALSE);

      if(c64_debug_level >= 10)
      {
         sUI numMsg = MLB_REG(MLB_MSGSTATUS_1) & 7; // mask NBOFMSGMB field
         
         printk(KERN_INFO "c64::mlb_send_to_dsp: %u unread msg on GPP-side.\n", numMsg);
      }
      
      /* Check FIFOSTATUS_0 and return EAGAIN if FIFO is still full */
      if(c64_debug_level >= 10)
      {
         sU32 numMsgQueued = MLB_REG(MLB_MSGSTATUS_0);
         //sU32 numMsgQueued = MLB_REG(MLB_MSGSTATUS_1);
         
         printk(KERN_INFO "c64::mlb_send_to_dsp: %u msg already in FIFO.\n", numMsgQueued);
      }
      
      /* Will be written to DSP_MLB_RAM_MSGSERIAL when loc_mlb_reset() is called */
      msg_serial++;
      
#ifdef HAVE_BROKEN_MAILBOX
      /* Workaround for buggy hardware (reading MESSAGE registers does not clear FIFO and
       *  once a message is sent, IRQs trigger continuously)
       *
       *  (note) loc_mlb_reset() will also store the current message serial (while waiting
       *          for the mailbox reset)
       */
      loc_mlb_reset();
      loc_mlb_reset_2();
#endif
      
      if(c64_debug_level >= 10)
      {
         sU32 val;

#ifdef PANDORA_VID2_MSG_HACK
         val = HACK_REG_READ(DSP_MLB_RAM_MSGSERIAL);
#else
         val = MSG_REG_READ(DSP_MLB_RAM_MSGSERIAL);      // 86000080
#endif /* PANDORA_VID2_MSG_HACK */
         printk(KERN_INFO "c64::mlb_send_to_dsp: msgSerial=0x%08x (msg_serial=0x%08x)\n", val, msg_serial);
         
         val = MSG_REG_READ(DSP_MLB_RAM_WATCHDOG);       // 860000B0
         printk(KERN_INFO "c64::mlb_send_to_dsp: dspIterCount=0x%08x\n", val);
         
         val = MSG_REG_READ(DSP_MLB_RAM_DSPIRQCOUNT);    // 860000B4
         printk(KERN_INFO "c64::mlb_send_to_dsp: dspIrqCount=0x%08x\n", val);
         
         val = MSG_REG_READ(DSP_MLB_RAM_DSPLASTREADMSG); // 860000B8
         printk(KERN_INFO "c64::mlb_send_to_dsp: dspLastReadMsg=0x%08x\n", val);
      }
      
#if 0
      /* Clears the FIFO but DSP still never receives NEWMSG interrupt, except after the very first msg sent.
       *  Resetting the mailbox seems to be the only working solution :/
       */
      for(;;)
      {
         sU32 numMsgQueued = MLB_REG(MLB_MSGSTATUS_0);
         
         if(numMsgQueued > 0)
         {
            printk(KERN_INFO "c64::mlb_send_to_dsp: %u msg still in FIFO. dspInbox=0x%08x\n", 
                   numMsgQueued,
                   mlb_dsp_inbox
                   );
            
            /* Clear DSP FIFO */
            mlb_dsp_inbox = MLB_REG(MLB_MESSAGE_0);
         }
         else
         {
            break;
         }
      }
#endif
      
      
      if(0 == (MLB_REG(MLB_FIFOSTATUS_0) & 1))
         //if(0 == (MLB_REG(MLB_FIFOSTATUS_1) & 1))
      {
         if(c64_debug_level >= 10)
         {
            printk(KERN_INFO "c64::mlb::mlb_send_to_dsp: sending msg. client=0x%08x dspCompId=0x%08x dspCmdId=0x%08x arg1=0x%08x arg2=0x%08x\n", 
                   (sU32)_client,
                   _msg->req.dsp_component_id,
                   _msg->req.dsp_cmd_id,
                   _msg->data[0].arg,
                   _msg->data[1].arg
                   );
         }

         /* Copy message into GPP<>DSP shared memory area and send to DSP */
         {
            volatile dsp_msg_t *dspMsg = (volatile dsp_msg_t*) MSG_ADDR(DSP_MLB_RAM_MSGTODSP);

            dspMsg->client = _client;  /* not used on DSP-side but read in dev_handle_reply() */

            if(NULL != _client)
            {
#if 0
               /* cache coherency issues */
               dspMsg->c64_msg = *_msg;
#else

#if 0
            // Abuse some registers
#define REGX  (0x48050580u)
               reg32_write(REGX, 0x12345678u); //  DSP_MLB_RAM_MSGTODSP + 0u, 
               printk(KERN_INFO " 0x%08x = 0x%08x\n", REGX, reg32_read(REGX));
#endif // 0

#ifdef PANDORA_VID2_MSG_HACK
               if(c64_debug_level >= 20)
               {
                  printk(KERN_INFO "writing msg serial 0x%08x\n", msg_serial);
               }

               HACK_REG_WRITE(DSP_MLB_RAM_MSGSERIAL, msg_serial);

               if(c64_debug_level >= 20)
               {
                  printk(KERN_INFO "wrote msg serial 0x%08x\n",
                         HACK_REG_READ(DSP_MLB_RAM_MSGSERIAL)
                         );
               }
#if 1
               HACK_REG_WRITE(DSP_MLB_RAM_MSGTODSP_COMP, 
                              (_msg->req.dsp_component_id) |        // lo
                              (((sU32)_msg->req.dsp_cmd_id) << 16)  // hi
                              );

               HACK_REG_WRITE(DSP_MLB_RAM_MSGTODSP_ARG1, _msg->data[0].arg);
               HACK_REG_WRITE(DSP_MLB_RAM_MSGTODSP_ARG2, _msg->data[1].arg);

               // for debugging purposes: unset return value
               HACK_REG_WRITE(DSP_MLB_RAM_MSGTOGPP_RET1, 0xCCccCCccu);
               HACK_REG_WRITE(DSP_MLB_RAM_MSGTOGPP_RET2, 0xDDddDDddu);
#endif 

#else
               /* Big problem: the memory is cached, even though ioremap_nocache() was used */
               MSG_REG_WRITE(DSP_MLB_RAM_MSGTODSP + 0u, (sU32)_client);
               MSG_REG_WRITE(DSP_MLB_RAM_MSGTODSP + 4u + 0u,
                             (_msg->req.dsp_component_id) |  // lo
                             (_msg->req.dsp_cmd_id << 16)    // hi
                             );
               MSG_REG_WRITE(DSP_MLB_RAM_MSGTODSP + 4u + 4u,
                             _msg->data[0].arg
                             );
               MSG_REG_WRITE(DSP_MLB_RAM_MSGTODSP + 4u + 8u,
                             _msg->data[1].arg
                             );
#endif /* PANDORA_VID2_MSG_HACK */

#endif
            }

#if 0
            flush_cache_all();
#endif
           
            /* Send message to DSP (triggers interrupt on DSP-side) */
            MLB_REG(MLB_MESSAGE_0) = (sU32) _msg;
            ////MLB_REG(MLB_MESSAGE_1) = (sU32) dspMsg;
         }

#if 0
         /* Testcode: try to read DSP L1DSRAM.
          *
          *  (note) does not work on the Pandora, maybe because the MMU is not setup for this ?
          *
          *  (note) this _does_ work on e.g. DRA44x.
          */
         {
            sU32 val;
            
            ////val = reg32_read(0x10f04000u);  // --> non-linefetch abort @0xe0cf0000
            ////printk(KERN_INFO "L1DSRAM[0] = 0x%08x\n", val);
         }
         
         //reg32_write(IVA2_IC_EVTSET_1, (55 & 31)); // force mailbox interrupt (event source 55)
         //////DSP_REG(0xB2000000u + IVA2_IC_EVTSET_1) = (55 & 31);
#endif
         
         /* Succeeded */
         ret = 0;
      }
      else
      {
         /* FIFO is still full, must retry later */

         atomic_set(&at_b_ready_to_send, S_TRUE);
         
         ret = -EAGAIN;
      }
   }
   else
   {
      /* Failed: still waiting for reply from DSP */

      ret = -EAGAIN;
   }

   return ret;
}


/*--------------------------------------------------------------------------- mlb_is_ready_to_send() */
sBool mlb_is_ready_to_send(void) {

   return (sBool) atomic_read(&at_b_ready_to_send);
}


/*--------------------------------------------------------------------------- mlb_debug_read() */
sU32 mlb_debug_read(void) {
   sU32 ret;

   ret = MSG_REG_READ(DSP_MLB_RAM_DSPDEBUG); // 860000BC
   ////printk(KERN_INFO "c64::mlb_debug_read: ret=0x%08x\n", val);

   return ret;
}


/*--------------------------------------------------------------------------- mlb_init() */
sBool mlb_init(void) {
   sBool ret;
   sU32 coreCmIdleSt;

   atomic_set(&at_b_ready_to_send, S_FALSE);

   /* Power up mailbox registers */
   reg32_bit_set(CORE_CM_ICLKEN1, 7);  // set EN_MAILBOXES bit

#if 1
   /* bad: (and can probably be removed) */
   {
      sUI i=0;

      for(i=0; i<40; i++)
      {
         udelay(1000);
      }
   }
#endif

   /* First msg serial sent to DSP will be 1 (inc'd in mlb_send_to_dsp()) */
   msg_serial = 0;

   /* Check whether mailbox regs can now be accessed */
   coreCmIdleSt = reg32_read(CORE_CM_IDLEST1);

#if 0
   if(c64_debug_level >= 20)
   {
      printk(KERN_INFO "c64::mlb_init: CM_IDLEST1_CORE=0x%08x\n",
             coreCmIdleSt
             );
   }
#endif

   /* check ST_MAILBOXES bit */
   if(0 == (coreCmIdleSt & 0x80))
   {
      msg_base_virt =
         (sU8 __iomem *)
         ioremap_nocache(MSG_BASE,
                         DSP_RESET_VECTOR_MSGSIZE
                         );

      if(NULL != msg_base_virt)
      {
         mlb_base_virt = (sU8 __iomem *) ioremap_nocache(MLB_BASE, 4096);
         
         if(NULL != mlb_base_virt)
         {
#ifdef PANDORA_VID2_MSG_HACK
            hack_base_virt =
            (sU8 __iomem *)
               ioremap_nocache(DSP_HACK_BASE,
                               32
                               );
            
            if(NULL != hack_base_virt)
#endif /* PANDORA_VID2_MSG_HACK */
            {
               sU32 mlbRev;
               
               mlbRev = MLB_REG(MLB_REVISION);
#if 1
               if(c64_debug_level >= 10)
               {
                  printk(KERN_INFO "c64::mlb_init: mlb revision=0x%08x\n", mlbRev); 
               }
#endif
               
               /* Disable mailbox interrupts and reset mailbox */
               loc_mlb_reset();
               
               
               /* Install interrupt handler for GPP mailbox interrupt */
               {
                  int r;
                  
                  r = request_irq(MLB_U0_MPU_IRQ,
                                  &irq_handler__mlb,
                                  0,//IRQF_DISABLED,      // fast interrupt handler, (see interrupt.h:IRQF_xxx)
                                  "c64-mlb",          // dev-name (for /proc/interrupts)
                                  NULL                // dev-id (ptr. to priv. data)
                                  );
                  
                  if(0 == r)
                  {
                     ////enable_irq(MLB_U0_MPU_IRQ);
                     
                     /* Enable mailbox interrupts */
                     loc_mlb_reset_2();
                     
                     
#if 0
                     printk(KERN_INFO "c64::mlb_init: unmasking IVA2_IRQ_10.\n");
                     /* Unmask IVA2_IRQ_10 (mailbox interrupt) */
                     ////reg32_write(0x01c21070u, (1u << 10)); // --> bus error
                     printk(KERN_INFO "c64::mlb_init:   -> done.\n");
#endif
                     
                     atomic_set(&at_b_ready_to_send, S_TRUE);
                     
                     /* Succeeded */
                     ret = S_TRUE;
                  }
                  else
                  {
                     /* Failed: request_irq() */
                     printk(KERN_ERR "c64::mlb_init: request_irq() failed.\n");
                     
#ifdef PANDORA_VID2_MSG_HACK
                     iounmap(hack_base_virt);
                     hack_base_virt = NULL;
#endif /* PANDORA_VID2_MSG_HACK */

                     iounmap(mlb_base_virt);
                     mlb_base_virt = NULL;
                    
                     iounmap(msg_base_virt);
                     msg_base_virt = NULL;
                     
                     ret = S_FALSE;
                  }
               }
            }
#ifdef PANDORA_VID2_MSG_HACK
            else
            {
               /* Failed to map VID2 scaler registers */
               printk(KERN_ERR "c64::mlb_init: ioremap_nocache() [hack] (VID2 scaler) failed.\n");
               
               iounmap(mlb_base_virt);
               mlb_base_virt = NULL;
               
               iounmap(msg_base_virt);
               msg_base_virt = NULL;
               
               ret = S_FALSE;
            }
#endif /* PANDORA_VID2_MSG_HACK */
         }
         else
         {
            /* Failed: ioremap_nocache() [mlb] */
            printk(KERN_ERR "c64::mlb_init: ioremap_nocache() [mlb] failed.\n");
            
            iounmap(msg_base_virt);
            msg_base_virt = NULL;
            
            ret = S_FALSE;
         }
      }
      else
      {
         /* Failed: ioremap_nocache() [msg] */
         printk(KERN_ERR "c64::mlb_init: ioremap_nocache() [msg] failed.\n");
         
         ret = S_FALSE;
      }
   }
   else
   {
      /* Failed to enable access to mailbox registers */
      printk(KERN_ERR "c64::mlb_init: mailbox regs cannot be accessed.\n");

      ret = S_FALSE;
   }

   return ret;
}


/*--------------------------------------------------------------------------- mlb_exit() */
void mlb_exit(void) {

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::mlb_exit: ENTER  (%u irqs handled so far)\n", mlb_irq_count);
   }

   atomic_set(&at_b_ready_to_send, S_FALSE);

   ////disable_irq(MLB_U0_MPU_IRQ);

   /* Disable GPP mailbox interrupts */
   MLB_REG(MLB_IRQENABLE_0) = 0;

   /* Disable DSP mailbox interrupts */
   MLB_REG(MLB_IRQENABLE_1) = 0;

   /* Uninstall interrupt handler */
   free_irq(MLB_U0_MPU_IRQ,
            NULL             // dev-id (ptr. to priv. data)
            );

   /* Unmap mailbox registers */
   iounmap(mlb_base_virt);
   mlb_base_virt = NULL;

#ifdef PANDORA_VID2_MSG_HACK
   /* Unmap hack (VID2 scaler) registers */
   iounmap(hack_base_virt);
   hack_base_virt = NULL;
#endif /* PANDORA_VID2_MSG_HACK */

   /* Unmap DSP<>GPP shared memory area */
   iounmap(msg_base_virt);
   msg_base_virt = NULL;

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::mlb_exit: LEAVE\n");
   }
}
/* ----
 * ---- file   : dsp_c64.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 10Sep2013, 11Sep2013, 15Sep2013
 * ----
 * ----
 */

#define __NO_VERSION__

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include "../include/types.h"
#include "../include/arch/dm3730/dm3730.h"

#include "kmod.h"
#include "reg.h"
#include "dsp_c64.h"


/*--------------------------------------------------------------------------- dsp_init() */
void dsp_init(void) {

   // Disable GPTIMER5 functional clock (EN_GPT5) (DSPBIOS timer)
   ////reg32_bit_clear(IVA2_PER_CM_FCLKEN, 6);
   
   // Disable GPTIMER5 interface clock (EN_GPT5)
   ////reg32_bit_clear(IVA2_PER_CM_ICLKEN, 6);
   
   // Disable automatic transition (clear bit 0, bits 1..31 are reserved)
   reg32_write(IVA2_CM_CLKSTCTRL, 0);
   
   // Disable IVA2_CLK (clear bit 0, bits 1..31 are reserved)
   reg32_write(IVA2_CM_FCLKEN, 0);
   
   // Reset video sequencer (RST3 bit) and DSP (RST1 bit) (do NOT reset MMU or the GPP OS will freeze!!)
   ////reg32_write(IVA2_PRM_RSTCTRL, 5);
   reg32_write(IVA2_PRM_RSTCTRL, 1 + 4);
   
   // Enable IVA2_CLK (set bit 1, bits 1..31 are reserved)
   reg32_write(IVA2_CM_FCLKEN, 1);
   
   {
      if(c64_debug_level >= 10)
      {
         printk(KERN_INFO "c64: disabling IVA2_MMU2..\n");
      }
      
      // Disable MMU
      reg32_bit_clear(IVA2_MMU2_CNTL, 1);
      
      if(c64_debug_level >= 10)
      {
         printk(KERN_INFO "c64: ..OK, IVA2_MMU2 is now disabled.\n");
      }
   }

   if(0)
   {
      sU32 val;
      val = reg32_read(0x48004D00);
      printk(KERN_INFO "val=0x%08x\n", val);
   }

   // Enable automatic transition (sets reserved bit 1 ??)
   reg32_write(IVA2_CM_CLKSTCTRL, 3);
}


/*--------------------------------------------------------------------------- dsp_exit() */
void dsp_exit(void) {
   // Disable automatic transition (clear bit 0, bits 1..31 are reserved)
   reg32_write(IVA2_CM_CLKSTCTRL, 0);
   
   // Disable IVA2_CLK (clear bit 0, bits 1..31 are reserved)
   reg32_write(IVA2_CM_FCLKEN, 0);

   // Enable automatic transition (sets reserved bit 1 ??)
   reg32_write(IVA2_CM_CLKSTCTRL, 3);
}
/* ----
 * ---- file   : dev.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 11Sep2013, 12Sep2013, 14Sep2013
 * ----
 * ----
 */

#define __NO_VERSION__

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "../include/types.h"
#include "../include/msg.h"

#include "kmod.h"
#include "mlb.h"
#include "dev.h"


/*--------------------------------------------------------------------------- clients */
typedef enum {
   STATE_UNUSED        = 0,  // client slot is free for allocation (open).
   STATE_IDLE          = 1,  // client is connected but there's no activity ATM.
   STATE_REQ_QUEUED    = 2,  // (todo) message is queued and will be sent to DSP ASAP.
   STATE_REPLY_PENDING = 3,  // waiting for reply from DSP.
   STATE_REPLY_AVAIL   = 4,  // reply from DSP has been recv'd but not picked up by the client, yet.
} dev_client_state_t;


typedef struct {
   atomic_t at_state;

   struct file *filep;

   c64_msg_t c64_msg;

} dev_client_t;


static dev_client_t clients[C64_DEV_MAX_CLIENTS];

static volatile dev_client_t *current_client; // != NULL while transmission in progress


static dev_client_t *loc_find_client_by_filep(struct file *_filep) {
   dev_client_t *ret;
   sUI i;

   ret = clients;

   for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
   {
      if(ret->filep == _filep)
      {
         if(STATE_UNUSED != atomic_read(&ret->at_state))
         {
            /* Succeeded */
            break;
         }
      }

      /* Try next client struct */
      ret++;
   }

   if(C64_DEV_MAX_CLIENTS == i)
   {
      /* Failed: client not found */
      ret = NULL;
   }

   return ret;
}


/*--------------------------------------------------------------------------- FOPS module vars */
static int     fops__open     (struct inode *_inode, struct file *_filep);
static int     fops__release  (struct inode *_inode, struct file *_filep);
static ssize_t fops__read     (struct file *_filep, char __user *_buf, size_t _count, loff_t *_pos);
static ssize_t fops__write    (struct file *_filep, const char __user *_buf, size_t _count, loff_t *_pos);
//static long    fops__unlocked_ioctl (struct file *_filep, unsigned int, unsigned long);

#define FOP(a) .a = fops__##a

static struct file_operations fops = {
   FOP(open),
   FOP(release),
   FOP(read),
   FOP(write),
};

/* (note) may temporarily exceed C64_DEV_MAX_CLIENTS */
static atomic_t at_num_open = ATOMIC_INIT(0);


/*--------------------------------------------------------------------------- fops__open */
int fops__open(struct inode *_inode, struct file *_filep) {
   int ret;

   if(C64_DEV_MINOR == iminor(_inode))
   {
      if(atomic_inc_and_test(&at_num_open) < C64_DEV_MAX_CLIENTS)
      {
         dev_client_t *cl = loc_find_client_by_filep(_filep);
         
         if(NULL == cl)
         {
            /* Find first unused client */
            sUI i;
            cl = clients;

            for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
            {
               if(STATE_UNUSED == atomic_read(&cl->at_state))
               {
                  /* Found unused client struct */
                  cl->filep = _filep;
                  atomic_set(&cl->at_state, STATE_IDLE);
                  break;
               }
               
               /* Try next client struct */
               cl++;
            }

            if(C64_DEV_MAX_CLIENTS != i)
            {
               /* Succeeded */
               if(c64_debug_level > 0)
               {
                  printk(KERN_INFO "c64::dev::open: client connected (filep=0x%p).\n", _filep);
               }

               ret = 0;
            }
            else
            {
               /* Failed: all clients in use */
               if(c64_debug_level > 0)
               {
                  printk(KERN_ERR "c64::dev::open: all clients in use.\n");
               }

               ret = -EMFILE;
            }
         }
         else
         {
            /* Failed: client not found (internal error) */
            printk(KERN_ERR "c64::dev::open: internal error: client (filep=0x%p) already exists.\n", _filep);

            ret = -EIO;
         }
      }
      else
      {
         /* Failed: maximum number of file handles exceeded */
         atomic_dec(&at_num_open);

         ret = -EMFILE;
      }
   }
   else
   {
      /* Failed: wrong minor device number */
      ret = -ENODEV;
      
   }

   return ret;
}


/*--------------------------------------------------------------------------- fops__release */
int fops__release(struct inode *_inode, struct file *_filep) {
   int ret;
   dev_client_t *cl;

   atomic_dec(&at_num_open);

   cl = loc_find_client_by_filep(_filep);

   if(NULL != cl)
   {
      cl->filep = NULL;
      atomic_set(&cl->at_state, STATE_UNUSED);

      if(c64_debug_level > 0)
      {
         printk(KERN_INFO "c64::dev::release: client disconnected (filep=0x%p).\n", _filep);
      }
   }
   else
   {
      /* Failed: client not found (internal error) */
      printk(KERN_ERR "c64::dev::release: internal error: client (filep=0x%p) not found.\n", _filep);
   }

   /* Succeeded */
   ret = 0;

   return ret;
}


/*--------------------------------------------------------------------------- fops__read */
ssize_t fops__read(struct file *_filep, char __user *_buf, size_t _count, loff_t *_pos) {
   ssize_t ret;

   (void)_pos;

   if(c64_debug_level >= 15)
   {
      printk(KERN_INFO "c64::dev::read: ENTER  count=%u\n", _count);
   }

   if(sizeof(c64_msg_t) == _count)
   {
      dev_client_t *cl = loc_find_client_by_filep(_filep);

      if(NULL != cl)
      {
         dev_client_state_t cState = atomic_read(&cl->at_state);

         if(STATE_REPLY_AVAIL == cState)
         {
            if(0 == copy_to_user(_buf, &cl->c64_msg, sizeof(c64_msg_t)))
            {
               /* Succeeded */
               if(c64_debug_level >= 15)
               {
                  printk(KERN_INFO "c64::dev:read: DSP reply picked up by application.\n");
               }

               atomic_set(&cl->at_state, STATE_IDLE);

               ret = sizeof(c64_msg_t);
            }
            else
            {
               /* Failed: copy_to_user() */
               printk(KERN_ERR "c64::dev::read: copy_to_user() failed.\n");

               return -EFAULT;
            }
         }
         else
         {
            if(STATE_REPLY_PENDING == cState)
            {
               /* Failed: no reply available, yet. */
               if(c64_debug_level >= 15)
               {
                  printk(KERN_ERR "c64::dev::read: reply still pending, cState=%d.\n", cState);
               }

               ret = -EAGAIN;
            }
            else
            {
               /* Failed: (warning) nothing to read. */
               if(c64_debug_level >= 15)
               {
                  printk(KERN_ERR "c64::dev::read: nothing to read (no pending reply).\n");
               }

               ret = 0;
            }
         }
      }
      else
      {
         /* Failed: client not found */
         if(c64_debug_level > 0)
         {
            printk(KERN_ERR "c64::dev::read: client (filep=0x%p) not found.\n", _filep);
         }

         ret = -EBADF;
      }
   }
   else if(4 == _count)
   {
      /* Read debug value */
      sU32 val = mlb_debug_read();

      if(0 == copy_to_user(_buf, &val, 4))
      {
         /* Succeeded */
         ret = 4;
      }
      else
      {
         /* Failed: copy_to_user() */
         return -EFAULT;
      }
   }
   else
   {
      /* Failed: invalid request size */
      if(c64_debug_level > 0)
      {
         printk(KERN_ERR "c64::dev::read: invalid size %d.\n", _count);
      }

      ret = -EINVAL;
   }

   if(c64_debug_level >= 15)
   {
      printk(KERN_INFO "c64::dev::read: LEAVE  ret=%d\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- fops__write */
ssize_t fops__write(struct file *_filep, const char __user *_buf, size_t _count, loff_t *_pos) {
   ssize_t ret;

   if(sizeof(c64_msg_t) == _count)
   {
      if(NULL == current_client)
      {
         dev_client_t *cl = loc_find_client_by_filep(_filep);
         
         if(NULL != cl)
         {
            if(0 == copy_from_user(&cl->c64_msg, _buf, sizeof(c64_msg_t)))
            {
               current_client = cl;
               
               atomic_set(&cl->at_state, STATE_REPLY_PENDING);

               ret = (ssize_t) mlb_send_to_dsp(cl, &cl->c64_msg);
               
               if(0 == ret)
               {
                  /* Succeeded */
                  ret = sizeof(c64_msg_t);
               }
               else
               {
                  /* Failed: mlb_send_to_dsp() */
                  if(c64_debug_level >= 15)
                  {
                     printk(KERN_ERR "c64::dev::write: mlb_send_to_dsp() failed (still busy).\n");
                  }
                  
                  current_client = NULL;

                  atomic_set(&cl->at_state, STATE_IDLE);
                  
                  ret = -EAGAIN;
               }
            }
            else
            {
               /* Failed: copy_from_user() */
               if(c64_debug_level > 0)
               {
                  printk(KERN_ERR "c64::dev::write: copy_from_user() failed.\n");
               }
               
               ret = -EFAULT;
            }
         }
         else
         {
            /* Failed: client not found */
            if(c64_debug_level > 0)
            {
               printk(KERN_ERR "c64::dev::write: client (filep=0x%p) not found.\n", _filep);
            }
            
            ret = -EBADF;
         }
      }
      else
      {
         /* Failed: transmission still in progress (current_client != NULL) */
         if(c64_debug_level >= 15)
         {
            printk(KERN_ERR "c64::dev::write: (current_client != NULL) (still busy).\n");
         }
         
         ret = -EAGAIN;
      }
   }
   else if(4 == _count)
   {
      /* Send test message to DSP, bypassing client handler */
      sU32 val;

      if(0 == copy_from_user(&val, _buf, 4))
      {
         current_client = NULL;

         ret = (ssize_t) mlb_send_to_dsp(NULL, (const c64_msg_t*) &val);

         if(0 == ret)
         {
            ret = 4;
         }
         else
         {
            /* Failed: mlb_send_to_dsp() */
            if(c64_debug_level > 0)
            {
               printk(KERN_ERR "c64::dev::write: mlb_send_to_dsp() failed (testmsg).\n");
            }
         }
      }
      else
      {
         /* Failed: copy_from_user() */
         if(c64_debug_level > 0)
         {
            printk(KERN_ERR "c64::dev::write: copy_from_user() (4) failed.\n");
         }
         
         ret = -EFAULT;
      }
   }
   else
   {
      /* Failed: invalid request size */
      if(c64_debug_level > 0)
      {
         printk(KERN_ERR "c64::dev::write: invalid size %d.\n", _count);
      }
      
      ret = -EINVAL;
   }

   return ret;
}


/*--------------------------------------------------------------------------- dev_handle_reply() */
void dev_handle_reply(const volatile dsp_msg_t *_msgTo,   /* message from GPP to DSP (for client handle) */
                      const volatile dsp_msg_t *_msgFrom  /* message from DSP to GPP */
                      ) {

   /* (note) currently called directly from mlb IRQ handler */
   
   if(c64_debug_level >= 10)
   {
      printk(KERN_INFO "c64::dev_handle_reply:: recv'd msg. client=0x%08x dspCompId=0x%08x dspCmdId=0x%08x ret1=0x%08x ret2=0x%08x\n", 
             (sU32)_msgTo->client,
             _msgTo->c64_msg.req.dsp_component_id,
             _msgTo->c64_msg.req.dsp_cmd_id,
             _msgFrom->c64_msg.data[0].ret,
             _msgFrom->c64_msg.data[1].ret
             );
   }

   if(NULL != current_client)
   {
      /* Integrity check: state */
      dev_client_state_t cState = (dev_client_state_t) atomic_read(&current_client->at_state);

      if(STATE_REPLY_PENDING == cState)
      {
         /* Integrity check: client ptr */
         if(_msgTo->client == current_client)
         {
            /* Copy message */
            current_client->c64_msg = _msgFrom->c64_msg;

            atomic_set(&current_client->at_state, STATE_REPLY_AVAIL);

            /* (note) another client may already be exchanging a message when
             *         this client picks up the DSP reply.
             */
            current_client = NULL;
         }
      }
      else
      {
         /* (note) in order to prevent kernel log spam in case sth goes terribly wrong,
          *         let's just print the error message in debug mode.
          */
         if(c64_debug_level > 0)
         {
            printk(KERN_ERR "c64::dev_handle_reply: unexpected client state %d.\n",
                   cState
                   );
         }
      }
   }
   else
   {
      /* (note) in order to prevent kernel log spam in case sth goes terribly wrong,
       *         let's just print the error message in debug mode.
       */
      if(c64_debug_level > 0)
      {
         printk(KERN_ERR "c64::dev_handle_reply: no current client!!\n");
      }
   }
   
}


/*--------------------------------------------------------------------------- dev_init() */
sBool dev_init(void) {
   sBool ret;
   int r;

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::dev_init: ENTER\n");
   }

   current_client = NULL;

   /* Register char device 
    *
    *  (todo) alloc major number dynamically and use script to create device node
    */
   r = register_chrdev(C64_DEV_MAJOR,
                       "c64",
                       &fops
                       );

   if(r >= 0)
   {
      /* Succeeded */
      if(c64_debug_level > 0)
      {
         printk(KERN_INFO "c64: registered char device major=%d minor=%d name=\"%s\"\n",
                C64_DEV_MAJOR,
                C64_DEV_MINOR,
                C64_DEV_NAME
                );
      }

      /* Initialize clients array */
      {
         sUI i;
         
         for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
         {
            atomic_t atInit = ATOMIC_INIT(STATE_UNUSED);
            clients[i].at_state = atInit; /* (todo) is this allowed ? */

            /* Next client */
         }
      }

      ret = S_TRUE;
   }
   else
   {
      /* Failed: register_chrdev() */
      printk(KERN_ERR "c64: failed to register char device major=%d name=\"%s\"\n",
             C64_DEV_MAJOR,
             C64_DEV_NAME
             );

      ret = S_FALSE;
   }

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::dev_init: LEAVE\n");
   }

   return ret;
}


/*--------------------------------------------------------------------------- dev_exit() */
void dev_exit(void) {

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::dev_exit: ENTER\n");
   }

   /* Unregister char device */
   unregister_chrdev(C64_DEV_MAJOR, C64_DEV_NAME);

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::dev_exit: LEAVE\n");
   }
}
/* ----
 * ---- file   : kmod.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 10Sep2013, 11Sep2013, 14Sep2013, 15Sep2013
 * ----
 * ----
 */

#undef __KERNEL__
#define __KERNEL__

#undef MODULE
#define MODULE defined

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/export.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include "../include/types.h"
#include "../include/dsp_common.h"
#include "../include/msg.h"

#include "kmod.h"
#include "reg.h"
#include "mlb.h"
#include "dsp_c64.h"
#include "dev.h"


/*--------------------------------------------------------------------------- global vars */
sUI c64_debug_level = 0;


/*--------------------------------------------------------------------------- module vars */


/*--------------------------------------------------------------------------- c64_init() */
static int __init c64_init(void) {
   int ret;

   if(c64_debug_level > 0)
   {
      printk(KERN_INFO "c64::c64_init: ENTER\n");
   }

   /* Put DSP into reset and disable MMU */
   dsp_init();

   /* Setup mailbox and interrupt handler */
   if(mlb_init())
   {
      /* Register "/dev/c64" char device */
      if(dev_init())
      {
         /* Succeeded */
         ret = 0;
      }
      else
      {
         /* Failed: dev_init() */
         mlb_exit();
         dsp_exit();

         ret = -1;
      }
   }
   else
   {
      /* Failed: mlb_init() */

      dsp_exit();

      ret = -1;
   }

   if(c64_debug_level > 0)
   {
      printk(KERN_INFO "c64::c64_init: LEAVE ret=%d\n", ret);
   }
   else
   {
      printk(KERN_INFO "c64: init ok. module build date: %s\n", __DATE__
#ifdef PANDORA_VID2_MSG_HACK
             "  (**VID2_MSG_HACK**)"
#endif /* PANDORA_VID2_MSG_HACK */
             );
   }

   return ret;
}


/*--------------------------------------------------------------------------- c64_exit() */
static void __exit c64_exit(void) {

   if(c64_debug_level > 0)
   {
      printk(KERN_INFO "c64::c64_exit: ENTER\n");
   }

   /* Unregister char device */
   dev_exit();

   /* Disable mailbox interrupt and unregister handler */
   mlb_exit();

   /* Power off IVA2 subsystem */
   dsp_exit();

   if(c64_debug_level >= 10)
   {
      printk(KERN_INFO "c64::c64_exit: LEAVE\n");
   }
}


/*--------------------------------------------------------------------------- c64_param_set_debug_level() */
static int c64_param_set_debug_level(const char *_val, struct kernel_param *_kp) {
   (void)_kp;

   sscanf(_val, "%u", &c64_debug_level);

   return 0;
}


/*--------------------------------------------------------------------------- module exports */
MODULE_AUTHOR("Bastian Spiegel");
MODULE_LICENSE("GPL");

MODULE_DESCRIPTION("TMS320C6000 init + ARM<>DSP messaging driver");
MODULE_SUPPORTED_DEVICE("omap3530, dm3730 (Open Pandora)");

////module_param(c64_debug_level, int, 0);
module_param_call(debug, &c64_param_set_debug_level, NULL, NULL, 0200);
MODULE_PARM_DESC(debug, "debug level (0=no debug output, 20=trace reg access)");

module_init(c64_init);
module_exit(c64_exit);
/* ----
 * ---- file   : dsp_config.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the LESSER GNU GENERAL PUBLIC LICENSE (LGPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING_LGPL for further information.
 * ----
 * ---- info   : DSP memory configuration. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 07Sep2013, 08Sep2013, 11Sep2013, 13Sep2013
 * ----
 * ----
 */

#ifndef C64_KBUILD
#include "../../../include/types.h"
#include "../../../include/dsp_common.h"
#include "../../../include/dsp_config.h"
#endif // C64_KBUILD


#define CFG_PANDORA_128M     // 0x86000000..87000000 (boot linux w/ mem=96M@0x80000000) ("classic" edition)
//                           //  (can also be used as a shared config that works for both classic and rebirth/1ghz editions
//                                by setting up a memory hole in the linux bootargs, i.e.
//                                mem=96M@0x80000000 mem=384M@0x88000000 (rebirth/1Ghz) resp.
//                                mem=96M@0x80000000 mem=127M@0x88000000 (classic)
//                                )

//#define CFG_PANDORA_512M   // 0x9E000000..A0000000 (boot linux w/ mem=480m) ("rebirth" and "1ghz" editions)



#if defined(CFG_PANDORA_512M)
dsp_config_t dsp_config = {
#error This dsp_config causes kernel crashes although no syscall (mmap) reports any error
   .reset_vector = {
      0x9E000000u,
      0x00000000u,
      0x00001000u
   },
   
#ifndef CCS
   .ram = {
      0x9E001000u,
      0x00000000u,
      0x00FCF000u
   },

   .pool = {
      0x9F000000u,
      0x00000000u,
      0x01000000u
   }
#endif /* CCS */
};


#elif defined(CFG_PANDORA_128M)
dsp_config_t dsp_config = {
#ifndef CCS
   .reset_vector = {
      0x86000000u,
      0x00000000u,
      0x00001000u
   },
#else
   {
      0x86000000u,
      0x00000000u,
      0x00001000u
   },
#endif /* CCS */
   
#ifndef CCS
   .ram = {
      0x86001000u,
      0x00000000u,
      0x00FCF000u
   },

   .pool = {
      0x87000000u,
      0x00000000u,
      0x01000000u
   }
#endif /* CCS */
};

#else
#error No dsp_config defined. Please edit "dsp_config.c".
#endif
