/* ----
 * ---- file   : dev.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 11Sep2013, 12Sep2013, 14Sep2013
 * ----
 * ----
 */

#define __NO_VERSION__

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "../include/types.h"
#include "../include/msg.h"

#include "kmod.h"
#include "mlb.h"
#include "dev.h"


/*--------------------------------------------------------------------------- clients */
typedef enum {
   STATE_UNUSED        = 0,  // client slot is free for allocation (open).
   STATE_IDLE          = 1,  // client is connected but there's no activity ATM.
   STATE_REQ_QUEUED    = 2,  // (todo) message is queued and will be sent to DSP ASAP.
   STATE_REPLY_PENDING = 3,  // waiting for reply from DSP.
   STATE_REPLY_AVAIL   = 4,  // reply from DSP has been recv'd but not picked up by the client, yet.
} dev_client_state_t;


typedef struct {
   atomic_t at_state;

   struct file *filep;

   c64_msg_t c64_msg;

} dev_client_t;


static dev_client_t clients[C64_DEV_MAX_CLIENTS];

static volatile dev_client_t *current_client; // != NULL while transmission in progress


static dev_client_t *loc_find_client_by_filep(struct file *_filep) {
   dev_client_t *ret;
   sUI i;

   ret = clients;

   for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
   {
      if(ret->filep == _filep)
      {
         if(STATE_UNUSED != atomic_read(&ret->at_state))
         {
            /* Succeeded */
            break;
         }
      }

      /* Try next client struct */
      ret++;
   }

   if(C64_DEV_MAX_CLIENTS == i)
   {
      /* Failed: client not found */
      ret = NULL;
   }

   return ret;
}


/*--------------------------------------------------------------------------- FOPS module vars */
static int     fops__open     (struct inode *_inode, struct file *_filep);
static int     fops__release  (struct inode *_inode, struct file *_filep);
static ssize_t fops__read     (struct file *_filep, char __user *_buf, size_t _count, loff_t *_pos);
static ssize_t fops__write    (struct file *_filep, const char __user *_buf, size_t _count, loff_t *_pos);
//static long    fops__unlocked_ioctl (struct file *_filep, unsigned int, unsigned long);

#define FOP(a) .a = fops__##a

static struct file_operations fops = {
   FOP(open),
   FOP(release),
   FOP(read),
   FOP(write),
};

/* (note) may temporarily exceed C64_DEV_MAX_CLIENTS */
static atomic_t at_num_open = ATOMIC_INIT(0);


/*--------------------------------------------------------------------------- fops__open */
int fops__open(struct inode *_inode, struct file *_filep) {
   int ret;

   if(C64_DEV_MINOR == iminor(_inode))
   {
      if(atomic_inc_and_test(&at_num_open) < C64_DEV_MAX_CLIENTS)
      {
         dev_client_t *cl = loc_find_client_by_filep(_filep);
         
         if(NULL == cl)
         {
            /* Find first unused client */
            sUI i;
            cl = clients;

            for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
            {
               if(STATE_UNUSED == atomic_read(&cl->at_state))
               {
                  /* Found unused client struct */
                  cl->filep = _filep;
                  atomic_set(&cl->at_state, STATE_IDLE);
                  break;
               }
               
               /* Try next client struct */
               cl++;
            }

            if(C64_DEV_MAX_CLIENTS != i)
            {
               /* Succeeded */
               if(c64_debug_level > 0)
               {
                  printk(KERN_INFO "c64::dev::open: client connected (filep=0x%p).\n", _filep);
               }

               ret = 0;
            }
            else
            {
               /* Failed: all clients in use */
               if(c64_debug_level > 0)
               {
                  printk(KERN_ERR "c64::dev::open: all clients in use.\n");
               }

               ret = -EMFILE;
            }
         }
         else
         {
            /* Failed: client not found (internal error) */
            printk(KERN_ERR "c64::dev::open: internal error: client (filep=0x%p) already exists.\n", _filep);

            ret = -EIO;
         }
      }
      else
      {
         /* Failed: maximum number of file handles exceeded */
         atomic_dec(&at_num_open);

         ret = -EMFILE;
      }
   }
   else
   {
      /* Failed: wrong minor device number */
      ret = -ENODEV;
      
   }

   return ret;
}


/*--------------------------------------------------------------------------- fops__release */
int fops__release(struct inode *_inode, struct file *_filep) {
   int ret;
   dev_client_t *cl;

   atomic_dec(&at_num_open);

   cl = loc_find_client_by_filep(_filep);

   if(NULL != cl)
   {
      cl->filep = NULL;
      atomic_set(&cl->at_state, STATE_UNUSED);

      if(c64_debug_level > 0)
      {
         printk(KERN_INFO "c64::dev::release: client disconnected (filep=0x%p).\n", _filep);
      }
   }
   else
   {
      /* Failed: client not found (internal error) */
      printk(KERN_ERR "c64::dev::release: internal error: client (filep=0x%p) not found.\n", _filep);
   }

   /* Succeeded */
   ret = 0;

   return ret;
}


/*--------------------------------------------------------------------------- fops__read */
ssize_t fops__read(struct file *_filep, char __user *_buf, size_t _count, loff_t *_pos) {
   ssize_t ret;

   (void)_pos;

   if(c64_debug_level >= 15)
   {
      printk(KERN_INFO "c64::dev::read: ENTER  count=%u\n", _count);
   }

   if(sizeof(c64_msg_t) == _count)
   {
      dev_client_t *cl = loc_find_client_by_filep(_filep);

      if(NULL != cl)
      {
         dev_client_state_t cState = atomic_read(&cl->at_state);

         if(STATE_REPLY_AVAIL == cState)
         {
            if(0 == copy_to_user(_buf, &cl->c64_msg, sizeof(c64_msg_t)))
            {
               /* Succeeded */
               if(c64_debug_level >= 15)
               {
                  printk(KERN_INFO "c64::dev:read: DSP reply picked up by application.\n");
               }

               atomic_set(&cl->at_state, STATE_IDLE);

               ret = sizeof(c64_msg_t);
            }
            else
            {
               /* Failed: copy_to_user() */
               printk(KERN_ERR "c64::dev::read: copy_to_user() failed.\n");

               return -EFAULT;
            }
         }
         else
         {
            if(STATE_REPLY_PENDING == cState)
            {
               /* Failed: no reply available, yet. */
               if(c64_debug_level >= 15)
               {
                  printk(KERN_ERR "c64::dev::read: reply still pending, cState=%d.\n", cState);
               }

               ret = -EAGAIN;
            }
            else
            {
               /* Failed: (warning) nothing to read. */
               if(c64_debug_level >= 15)
               {
                  printk(KERN_ERR "c64::dev::read: nothing to read (no pending reply).\n");
               }

               ret = 0;
            }
         }
      }
      else
      {
         /* Failed: client not found */
         if(c64_debug_level > 0)
         {
            printk(KERN_ERR "c64::dev::read: client (filep=0x%p) not found.\n", _filep);
         }

         ret = -EBADF;
      }
   }
   else if(4 == _count)
   {
      /* Read debug value */
      sU32 val = mlb_debug_read();

      if(0 == copy_to_user(_buf, &val, 4))
      {
         /* Succeeded */
         ret = 4;
      }
      else
      {
         /* Failed: copy_to_user() */
         return -EFAULT;
      }
   }
   else
   {
      /* Failed: invalid request size */
      if(c64_debug_level > 0)
      {
         printk(KERN_ERR "c64::dev::read: invalid size %d.\n", _count);
      }

      ret = -EINVAL;
   }

   if(c64_debug_level >= 15)
   {
      printk(KERN_INFO "c64::dev::read: LEAVE  ret=%d\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- fops__write */
ssize_t fops__write(struct file *_filep, const char __user *_buf, size_t _count, loff_t *_pos) {
   ssize_t ret;

   if(sizeof(c64_msg_t) == _count)
   {
      if(NULL == current_client)
      {
         dev_client_t *cl = loc_find_client_by_filep(_filep);
         
         if(NULL != cl)
         {
            if(0 == copy_from_user(&cl->c64_msg, _buf, sizeof(c64_msg_t)))
            {
               current_client = cl;
               
               atomic_set(&cl->at_state, STATE_REPLY_PENDING);

               ret = (ssize_t) mlb_send_to_dsp(cl, &cl->c64_msg);
               
               if(0 == ret)
               {
                  /* Succeeded */
                  ret = sizeof(c64_msg_t);
               }
               else
               {
                  /* Failed: mlb_send_to_dsp() */
                  if(c64_debug_level >= 15)
                  {
                     printk(KERN_ERR "c64::dev::write: mlb_send_to_dsp() failed (still busy).\n");
                  }
                  
                  current_client = NULL;

                  atomic_set(&cl->at_state, STATE_IDLE);
                  
                  ret = -EAGAIN;
               }
            }
            else
            {
               /* Failed: copy_from_user() */
               if(c64_debug_level > 0)
               {
                  printk(KERN_ERR "c64::dev::write: copy_from_user() failed.\n");
               }
               
               ret = -EFAULT;
            }
         }
         else
         {
            /* Failed: client not found */
            if(c64_debug_level > 0)
            {
               printk(KERN_ERR "c64::dev::write: client (filep=0x%p) not found.\n", _filep);
            }
            
            ret = -EBADF;
         }
      }
      else
      {
         /* Failed: transmission still in progress (current_client != NULL) */
         if(c64_debug_level >= 15)
         {
            printk(KERN_ERR "c64::dev::write: (current_client != NULL) (still busy).\n");
         }
         
         ret = -EAGAIN;
      }
   }
   else if(4 == _count)
   {
      /* Send test message to DSP, bypassing client handler */
      sU32 val;

      if(0 == copy_from_user(&val, _buf, 4))
      {
         current_client = NULL;

         ret = (ssize_t) mlb_send_to_dsp(NULL, (const c64_msg_t*) &val);

         if(0 == ret)
         {
            ret = 4;
         }
         else
         {
            /* Failed: mlb_send_to_dsp() */
            if(c64_debug_level > 0)
            {
               printk(KERN_ERR "c64::dev::write: mlb_send_to_dsp() failed (testmsg).\n");
            }
         }
      }
      else
      {
         /* Failed: copy_from_user() */
         if(c64_debug_level > 0)
         {
            printk(KERN_ERR "c64::dev::write: copy_from_user() (4) failed.\n");
         }
         
         ret = -EFAULT;
      }
   }
   else
   {
      /* Failed: invalid request size */
      if(c64_debug_level > 0)
      {
         printk(KERN_ERR "c64::dev::write: invalid size %d.\n", _count);
      }
      
      ret = -EINVAL;
   }

   return ret;
}


/*--------------------------------------------------------------------------- dev_handle_reply() */
void dev_handle_reply(const volatile dsp_msg_t *_msgTo,   /* message from GPP to DSP (for client handle) */
                      const volatile dsp_msg_t *_msgFrom  /* message from DSP to GPP */
                      ) {

   /* (note) currently called directly from mlb IRQ handler */
   
   if(c64_debug_level >= 10)
   {
      printk(KERN_INFO "c64::dev_handle_reply:: recv'd msg. client=0x%08x dspCompId=0x%08x dspCmdId=0x%08x ret1=0x%08x ret2=0x%08x\n", 
             (sU32)_msgTo->client,
             _msgTo->c64_msg.req.dsp_component_id,
             _msgTo->c64_msg.req.dsp_cmd_id,
             _msgFrom->c64_msg.data[0].ret,
             _msgFrom->c64_msg.data[1].ret
             );
   }

   if(NULL != current_client)
   {
      /* Integrity check: state */
      dev_client_state_t cState = (dev_client_state_t) atomic_read(&current_client->at_state);

      if(STATE_REPLY_PENDING == cState)
      {
         /* Integrity check: client ptr */
         if(_msgTo->client == current_client)
         {
            /* Copy message */
            current_client->c64_msg = _msgFrom->c64_msg;

            atomic_set(&current_client->at_state, STATE_REPLY_AVAIL);

            /* (note) another client may already be exchanging a message when
             *         this client picks up the DSP reply.
             */
            current_client = NULL;
         }
      }
      else
      {
         /* (note) in order to prevent kernel log spam in case sth goes terribly wrong,
          *         let's just print the error message in debug mode.
          */
         if(c64_debug_level > 0)
         {
            printk(KERN_ERR "c64::dev_handle_reply: unexpected client state %d.\n",
                   cState
                   );
         }
      }
   }
   else
   {
      /* (note) in order to prevent kernel log spam in case sth goes terribly wrong,
       *         let's just print the error message in debug mode.
       */
      if(c64_debug_level > 0)
      {
         printk(KERN_ERR "c64::dev_handle_reply: no current client!!\n");
      }
   }
   
}


/*--------------------------------------------------------------------------- dev_init() */
sBool dev_init(void) {
   sBool ret;
   int r;

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::dev_init: ENTER\n");
   }

   current_client = NULL;

   /* Register char device 
    *
    *  (todo) alloc major number dynamically and use script to create device node
    */
   r = register_chrdev(C64_DEV_MAJOR,
                       "c64",
                       &fops
                       );

   if(r >= 0)
   {
      /* Succeeded */
      if(c64_debug_level > 0)
      {
         printk(KERN_INFO "c64: registered char device major=%d minor=%d name=\"%s\"\n",
                C64_DEV_MAJOR,
                C64_DEV_MINOR,
                C64_DEV_NAME
                );
      }

      /* Initialize clients array */
      {
         sUI i;
         
         for(i=0; i<C64_DEV_MAX_CLIENTS; i++)
         {
            atomic_t atInit = ATOMIC_INIT(STATE_UNUSED);
            clients[i].at_state = atInit; /* (todo) is this allowed ? */

            /* Next client */
         }
      }

      ret = S_TRUE;
   }
   else
   {
      /* Failed: register_chrdev() */
      printk(KERN_ERR "c64: failed to register char device major=%d name=\"%s\"\n",
             C64_DEV_MAJOR,
             C64_DEV_NAME
             );

      ret = S_FALSE;
   }

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::dev_init: LEAVE\n");
   }

   return ret;
}


/*--------------------------------------------------------------------------- dev_exit() */
void dev_exit(void) {

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::dev_exit: ENTER\n");
   }

   /* Unregister char device */
   unregister_chrdev(C64_DEV_MAJOR, C64_DEV_NAME);

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::dev_exit: LEAVE\n");
   }
}
