/* ----
 * ---- file   : mlb.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 10Sep2013, 11Sep2013, 14Sep2013, 15Sep2013
 * ----
 * ----
 */

#define HAVE_BROKEN_MAILBOX defined


#define __NO_VERSION__

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <linux/../../arch/arm/include/asm/cacheflush.h>


#include "../include/types.h"
#include "../include/msg.h"
#include "../include/dsp_common.h"
#define C64_KBUILD defined
#include "../include/dsp_config.h"

#include "../include/arch/dm3730/dm3730.h"

#include "kmod.h"
#include "reg.h"
#include "dev.h"
#include "mlb.h"

////extern void omap_writel(u32 v, u32 pa); /* (todo) use proper include file */
////extern u32  omap_readl(u32 pa); /* (todo) use proper include file */


#define MLB_MSG_POOL_SIZE  (RESET_VECTOR_MSGSIZE / sizeof(dsp_msg_t))  // => 248


/*--------------------------------------------------------------------------- module vars */
static volatile sU32 mlb_irq_count = 0;
static volatile sU32 mlb_bh_count = 0;

static volatile sU32 mlb_num_recvd = 0;
static volatile sU32 mlb_num_sent = 0;  // == num revd by the DSP

static sU8 __iomem *mlb_base_virt;  /* also see "dsp_common.h" */

#define MLB_REG(a)  (*(volatile sU32*)( ((a) - MLB_BASE) + mlb_base_virt ))

static volatile sU32 mlb_inbox[MLB_FIFO_SIZE]; /* written by the IRQ handler. actually only one message. */

////static volatile sU32 mlb_dsp_inbox; // workaround for FIFO clear fail on DSP-side

static sU8 __iomem *msg_base_virt;

#define MSG_BASE  (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE)
#define MSG_ADDR(a) (((a) - MSG_BASE) + msg_base_virt)
#define MSG_REG(a)  (*(volatile sU32 __iomem*)( MSG_ADDR(a) ))


/* (note) calling ioremap_nocache() / iounmap() in interrupt handler freezes the OS */
//#define MSG_REG_READ(a) reg32_read(a)
//#define MSG_REG_WRITE(a, b) reg32_write(a, b)

// crashes:
/////#define MSG_REG_READ(a) omap_readl(a)
/////#define MSG_REG_WRITE(a, b) omap_writel(b, a)

// works but access still is cached (DSP sometimes sees wrong data):
#define MSG_REG_READ(a) __raw_readl((const volatile void __iomem*)(MSG_ADDR(a)))
#define MSG_REG_WRITE(a, b) __raw_writel(b, (volatile void __iomem*)MSG_ADDR(a))


//#define MSG_REG_READ(a) MSG_REG(a)
//#define MSG_REG_WRITE(a, b) MSG_REG(a) = b


#ifdef PANDORA_VID2_MSG_HACK
static sU8 __iomem *hack_base_virt;
#define HACK_ADDR(a) (((a) - DSP_HACK_BASE) + hack_base_virt)
#define HACK_REG(a)  (*(volatile sU32 __iomem*)( HACK_ADDR(a) ))

#define HACK_REG_READ(a) __raw_readl((const volatile void __iomem*)(HACK_ADDR(a)))
#define HACK_REG_WRITE(a, b) __raw_writel(b, (volatile void __iomem*)HACK_ADDR(a))
#endif /* PANDORA_VID2_MSG_HACK */


////static volatile sUI msg_idx;  // current send/recv message index

static sU32 msg_serial;  /* required for buggy HW workaround so the DSP can identify unique messages */

static atomic_t at_b_ready_to_send = ATOMIC_INIT(S_FALSE);  /* false while waiting for reply from DSP */


/*--------------------------------------------------------------------------- loc_mlb_reset() */
static void loc_mlb_reset(void) {

   /* Disable GPP mailbox interrupts */
   MLB_REG(MLB_IRQENABLE_0) = 0;
   
   /* Disable DSP mailbox interrupts */
   MLB_REG(MLB_IRQENABLE_1) = 0;

   /* Mailbox software reset + set SIDLEMODE to "smart-idle" and autoidle to
    *  automatic interface clocking (based on L4-core interface activity)
    */
   MLB_REG(MLB_SYSCONFIG) = 
      (MLB_REG(MLB_SYSCONFIG) & ~((1u << 3) | (1u << 4))) |
      (2u << 3) |  // SIDLEMODE
      (1u << 0) |  // AUTOIDLE
      (1u << 1) ;  // SOFTRESET

   /* Store message serial so the DSP can distinguish between interrupt spam
    *  (on my HW, it always receives at least two interrupts, even with the workarounds)
    *  and "real" messages.
    */
   if(c64_debug_level >= 20)
   {
      printk(KERN_DEBUG "writing msg_serial=0x%08x to 0x%08x\n", msg_serial, DSP_MLB_RAM_MSGSERIAL);
   }

#ifdef PANDORA_VID2_MSG_HACK
   if(c64_debug_level >= 20)
   {
      printk(KERN_DEBUG "write msgserial hack_base_virt=0x%08x, MSGSERIAL_phys=0x%08x MSGSERIAL_virt=0x%08x\n",
             (sU32)hack_base_virt,
             DSP_MLB_RAM_MSGSERIAL,
             (sU32)HACK_ADDR(DSP_MLB_RAM_MSGSERIAL)
             );
   }
   HACK_REG_WRITE(DSP_MLB_RAM_MSGSERIAL, msg_serial);
#else
   MSG_REG_WRITE(DSP_MLB_RAM_MSGSERIAL, msg_serial);
#endif /* PANDORA_VID2_MSG_HACK */

   /* Wait for mailbox software reset */
   while(1 != MLB_REG(MLB_SYSSTATUS))
   {
      /* Wait for RESETDONE bit */
   }
}


/*--------------------------------------------------------------------------- loc_mlb_reset_2() */
static void loc_mlb_reset_2(void) {
   /* GPP-side:
    *  Setup mailbox 0 for sending messages from the GPP to the DSP, and
    *        mailbox 1 for receiving messages from the DSP
    */
#if 1
   MLB_REG(MLB_IRQENABLE_0) = 
      (
         //(1u << 1) |  // Generate IRQ when msg was delivered to the DSP and FIFO slot becomes avail.
         (1u << 2)    // set NEWMSGENABLEUUMB1 --> Generate IRQ when msg was received from the DSP
       );
#else
   MLB_REG(MLB_IRQENABLE_1) = 
      (
         //(1u << 3) |  // Generate IRQ when msg was delivered to the GPP and FIFO slot becomes avail.
         (1u << 0)    // set NEWMSGENABLEUUMB0 --> Generate IRQ when msg was received from the GPP
       );
#endif
   
   
#if 1
   /* DSP-side:
    *  Setup mailbox 0 for receiving messages from the DSP to the GPP, and
    *        mailbox 1 for sending messages to the GPP
    */
#if 1
   MLB_REG(MLB_IRQENABLE_1) = 
      (
         //(1u << 3) |  // Generate IRQ when msg was delivered to the GPP and FIFO slot becomes avail.
         (1u << 0)    // set NEWMSGENABLEUUMB0 --> Generate IRQ when msg was received from the GPP
       );
#else
   MLB_REG(MLB_IRQENABLE_0) = 
      (
         //(1u << 1) |  // Generate IRQ when msg was delivered to the GPP and FIFO slot becomes avail.
         (1u << 2)    // set NEWMSGENABLEUUMB1 --> Generate IRQ when msg was received from the GPP
       );
#endif
#endif
}


/*--------------------------------------------------------------------------- bh_handler__mlb */
static void bh_handler__mlb(unsigned long _userData) {
   /* Bottom half IRQ handler (not used ATM, have to figured out why tasklet_schedule() freezes the OS.
       Probably not that important, anyway.
    */
   (void)_userData;

   mlb_bh_count++;

   ////wake_up_interruptible(&mlb_queue);
}

static DECLARE_TASKLET(c64_mlb_tasklet, bh_handler__mlb, 0);


/*--------------------------------------------------------------------------- irq_handler__mlb */
static irqreturn_t irq_handler__mlb(int _irq,
                                    void * _devId // priv data ptr (NULL)
                                    ) {
   /* Top half IRQ handler (actually does the entire IRQ handling ATM) */

   mlb_irq_count++;

   if(c64_debug_level >= 20)
   {
      printk(KERN_INFO "irq_handler_mlb\n");
   }

   /* Read messages */
   {
      sUI numMsg = MLB_REG(MLB_MSGSTATUS_1) & 7; // mask NBOFMSGMB field
      //sUI numMsg = MLB_REG(MLB_MSGSTATUS_0) & 7; // mask NBOFMSGMB field

      if(numMsg > 0)
      {
         if(numMsg <= MLB_FIFO_SIZE)  // sanity check
         {
            sUI fifoIdx = 0;

            while(fifoIdx < numMsg)
            {
               mlb_inbox[fifoIdx] = MLB_REG(MLB_MESSAGE_1);
               //mlb_inbox[fifoIdx] = MLB_REG(MLB_MESSAGE_0);

               //if(mlb_num_recvd < 10) // don't flood the klog
               {
                  mlb_num_recvd++;

                  if(c64_debug_level >= 16)
                  {
                     /* Uhh.. printk in an IRQ handler.. ;) */ 
                     printk(KERN_DEBUG "c64: mlb_inbox[%u/%u] = 0x%08x\n",
                            (fifoIdx + 1),
                            numMsg,
                            mlb_inbox[fifoIdx]
                            );
                  }
               }
               
               // Next message
               fifoIdx++;
            }
         }
         /* else: numMsg exceeds FIFO size */
      }
      /* else: no messages to read */
   }

   /* Acknowledge interrupt */
   MLB_REG(MLB_IRQSTATUS_0) = (1u << 2);  // clear NEWMSGSTATUSUUMB1
   //MLB_REG(MLB_IRQSTATUS_0) = 15;  // clear all bits (no change, still getting interrupt spam)
   ////MLB_REG(MLB_IRQSTATUS_1) |= (1 << 0);  // clear NEWMSGSTATUSUUMB0
   
   ////tasklet_schedule(&c64_mlb_tasklet); // freezes system

   /* Disable mailbox interrupts to prevent interrupt spam
    *  (workaround for buggy hardware)
    */
   MLB_REG(MLB_IRQENABLE_0) = 0;

   /* Copy DSP reply to client structure */
   {
#if 0
      dev_handle_reply((const volatile dsp_msg_t *)MSG_ADDR(DSP_MLB_RAM_MSGTODSP), /* for client handle */
                       (const volatile dsp_msg_t *)MSG_ADDR(DSP_MLB_RAM_MSGTOGPP)
                       );
#else
      dsp_msg_t reply;

#if 0
      flush_cache_all();
#endif

#ifdef PANDORA_VID2_MSG_HACK
      /* also see "dsp_commmon.h" */
      reply.c64_msg.data[0].ret = HACK_REG_READ(DSP_MLB_RAM_MSGTOGPP_RET1);
      reply.c64_msg.data[1].ret = HACK_REG_READ(DSP_MLB_RAM_MSGTOGPP_RET2);
#else
      /* (note) calling ioremap_nocache() / iounmap() in interrupt handler freezes the OS 
       * (note) this does not work occasionally since the memory seems cached, after all.
       */
      reply.c64_msg.data[0].ret = MSG_REG(DSP_MLB_RAM_MSGTOGPP + 4u + 4u);
      reply.c64_msg.data[1].ret = MSG_REG(DSP_MLB_RAM_MSGTOGPP + 4u + 8u);

#endif /* PANDORA_VID2_MSG_HACK */

      dev_handle_reply((const volatile dsp_msg_t *)MSG_ADDR(DSP_MLB_RAM_MSGTODSP),
                       &reply
                       );

#endif
   }

   /* Ready to send the next message */
   atomic_set(&at_b_ready_to_send, S_TRUE);

   return IRQ_HANDLED;
}


/*--------------------------------------------------------------------------- mlb_send_to_dsp() */
int mlb_send_to_dsp(void *_client, const c64_msg_t *_msg) {
   int ret;

   if(S_TRUE == atomic_read(&at_b_ready_to_send) || (NULL == _client)/*force testmsg*/)
   {
      atomic_set(&at_b_ready_to_send, S_FALSE);

      if(c64_debug_level >= 10)
      {
         sUI numMsg = MLB_REG(MLB_MSGSTATUS_1) & 7; // mask NBOFMSGMB field
         
         printk(KERN_INFO "c64::mlb_send_to_dsp: %u unread msg on GPP-side.\n", numMsg);
      }
      
      /* Check FIFOSTATUS_0 and return EAGAIN if FIFO is still full */
      if(c64_debug_level >= 10)
      {
         sU32 numMsgQueued = MLB_REG(MLB_MSGSTATUS_0);
         //sU32 numMsgQueued = MLB_REG(MLB_MSGSTATUS_1);
         
         printk(KERN_INFO "c64::mlb_send_to_dsp: %u msg already in FIFO.\n", numMsgQueued);
      }
      
      /* Will be written to DSP_MLB_RAM_MSGSERIAL when loc_mlb_reset() is called */
      msg_serial++;
      
#ifdef HAVE_BROKEN_MAILBOX
      /* Workaround for buggy hardware (reading MESSAGE registers does not clear FIFO and
       *  once a message is sent, IRQs trigger continuously)
       *
       *  (note) loc_mlb_reset() will also store the current message serial (while waiting
       *          for the mailbox reset)
       */
      loc_mlb_reset();
      loc_mlb_reset_2();
#endif
      
      if(c64_debug_level >= 10)
      {
         sU32 val;

#ifdef PANDORA_VID2_MSG_HACK
         val = HACK_REG_READ(DSP_MLB_RAM_MSGSERIAL);
#else
         val = MSG_REG_READ(DSP_MLB_RAM_MSGSERIAL);      // 86000080
#endif /* PANDORA_VID2_MSG_HACK */
         printk(KERN_INFO "c64::mlb_send_to_dsp: msgSerial=0x%08x (msg_serial=0x%08x)\n", val, msg_serial);
         
         val = MSG_REG_READ(DSP_MLB_RAM_WATCHDOG);       // 860000B0
         printk(KERN_INFO "c64::mlb_send_to_dsp: dspIterCount=0x%08x\n", val);
         
         val = MSG_REG_READ(DSP_MLB_RAM_DSPIRQCOUNT);    // 860000B4
         printk(KERN_INFO "c64::mlb_send_to_dsp: dspIrqCount=0x%08x\n", val);
         
         val = MSG_REG_READ(DSP_MLB_RAM_DSPLASTREADMSG); // 860000B8
         printk(KERN_INFO "c64::mlb_send_to_dsp: dspLastReadMsg=0x%08x\n", val);
      }
      
#if 0
      /* Clears the FIFO but DSP still never receives NEWMSG interrupt, except after the very first msg sent.
       *  Resetting the mailbox seems to be the only working solution :/
       */
      for(;;)
      {
         sU32 numMsgQueued = MLB_REG(MLB_MSGSTATUS_0);
         
         if(numMsgQueued > 0)
         {
            printk(KERN_INFO "c64::mlb_send_to_dsp: %u msg still in FIFO. dspInbox=0x%08x\n", 
                   numMsgQueued,
                   mlb_dsp_inbox
                   );
            
            /* Clear DSP FIFO */
            mlb_dsp_inbox = MLB_REG(MLB_MESSAGE_0);
         }
         else
         {
            break;
         }
      }
#endif
      
      
      if(0 == (MLB_REG(MLB_FIFOSTATUS_0) & 1))
         //if(0 == (MLB_REG(MLB_FIFOSTATUS_1) & 1))
      {
         if(c64_debug_level >= 10)
         {
            printk(KERN_INFO "c64::mlb::mlb_send_to_dsp: sending msg. client=0x%08x dspCompId=0x%08x dspCmdId=0x%08x arg1=0x%08x arg2=0x%08x\n", 
                   (sU32)_client,
                   _msg->req.dsp_component_id,
                   _msg->req.dsp_cmd_id,
                   _msg->data[0].arg,
                   _msg->data[1].arg
                   );
         }

         /* Copy message into GPP<>DSP shared memory area and send to DSP */
         {
            volatile dsp_msg_t *dspMsg = (volatile dsp_msg_t*) MSG_ADDR(DSP_MLB_RAM_MSGTODSP);

            dspMsg->client = _client;  /* not used on DSP-side but read in dev_handle_reply() */

            if(NULL != _client)
            {
#if 0
               /* cache coherency issues */
               dspMsg->c64_msg = *_msg;
#else

#if 0
            // Abuse some registers
#define REGX  (0x48050580u)
               reg32_write(REGX, 0x12345678u); //  DSP_MLB_RAM_MSGTODSP + 0u, 
               printk(KERN_INFO " 0x%08x = 0x%08x\n", REGX, reg32_read(REGX));
#endif // 0

#ifdef PANDORA_VID2_MSG_HACK
               if(c64_debug_level >= 20)
               {
                  printk(KERN_INFO "writing msg serial 0x%08x\n", msg_serial);
               }

               HACK_REG_WRITE(DSP_MLB_RAM_MSGSERIAL, msg_serial);

               if(c64_debug_level >= 20)
               {
                  printk(KERN_INFO "wrote msg serial 0x%08x\n",
                         HACK_REG_READ(DSP_MLB_RAM_MSGSERIAL)
                         );
               }
#if 1
               HACK_REG_WRITE(DSP_MLB_RAM_MSGTODSP_COMP, 
                              (_msg->req.dsp_component_id) |        // lo
                              (((sU32)_msg->req.dsp_cmd_id) << 16)  // hi
                              );

               HACK_REG_WRITE(DSP_MLB_RAM_MSGTODSP_ARG1, _msg->data[0].arg);
               HACK_REG_WRITE(DSP_MLB_RAM_MSGTODSP_ARG2, _msg->data[1].arg);

               // for debugging purposes: unset return value
               HACK_REG_WRITE(DSP_MLB_RAM_MSGTOGPP_RET1, 0xCCccCCccu);
               HACK_REG_WRITE(DSP_MLB_RAM_MSGTOGPP_RET2, 0xDDddDDddu);
#endif 

#else
               /* Big problem: the memory is cached, even though ioremap_nocache() was used */
               MSG_REG_WRITE(DSP_MLB_RAM_MSGTODSP + 0u, (sU32)_client);
               MSG_REG_WRITE(DSP_MLB_RAM_MSGTODSP + 4u + 0u,
                             (_msg->req.dsp_component_id) |  // lo
                             (_msg->req.dsp_cmd_id << 16)    // hi
                             );
               MSG_REG_WRITE(DSP_MLB_RAM_MSGTODSP + 4u + 4u,
                             _msg->data[0].arg
                             );
               MSG_REG_WRITE(DSP_MLB_RAM_MSGTODSP + 4u + 8u,
                             _msg->data[1].arg
                             );
#endif /* PANDORA_VID2_MSG_HACK */

#endif
            }

#if 0
            flush_cache_all();
#endif
           
            /* Send message to DSP (triggers interrupt on DSP-side) */
            MLB_REG(MLB_MESSAGE_0) = (sU32) _msg;
            ////MLB_REG(MLB_MESSAGE_1) = (sU32) dspMsg;
         }

#if 0
         /* Testcode: try to read DSP L1DSRAM.
          *
          *  (note) does not work on the Pandora, maybe because the MMU is not setup for this ?
          *
          *  (note) this _does_ work on e.g. DRA44x.
          */
         {
            sU32 val;
            
            ////val = reg32_read(0x10f04000u);  // --> non-linefetch abort @0xe0cf0000
            ////printk(KERN_INFO "L1DSRAM[0] = 0x%08x\n", val);
         }
         
         //reg32_write(IVA2_IC_EVTSET_1, (55 & 31)); // force mailbox interrupt (event source 55)
         //////DSP_REG(0xB2000000u + IVA2_IC_EVTSET_1) = (55 & 31);
#endif
         
         /* Succeeded */
         ret = 0;
      }
      else
      {
         /* FIFO is still full, must retry later */

         atomic_set(&at_b_ready_to_send, S_TRUE);
         
         ret = -EAGAIN;
      }
   }
   else
   {
      /* Failed: still waiting for reply from DSP */

      ret = -EAGAIN;
   }

   return ret;
}


/*--------------------------------------------------------------------------- mlb_is_ready_to_send() */
sBool mlb_is_ready_to_send(void) {

   return (sBool) atomic_read(&at_b_ready_to_send);
}


/*--------------------------------------------------------------------------- mlb_debug_read() */
sU32 mlb_debug_read(void) {
   sU32 ret;

   ret = MSG_REG_READ(DSP_MLB_RAM_DSPDEBUG); // 860000BC
   ////printk(KERN_INFO "c64::mlb_debug_read: ret=0x%08x\n", val);

   return ret;
}


/*--------------------------------------------------------------------------- mlb_init() */
sBool mlb_init(void) {
   sBool ret;
   sU32 coreCmIdleSt;

   atomic_set(&at_b_ready_to_send, S_FALSE);

   /* Power up mailbox registers */
   reg32_bit_set(CORE_CM_ICLKEN1, 7);  // set EN_MAILBOXES bit

#if 1
   /* bad: (and can probably be removed) */
   {
      sUI i=0;

      for(i=0; i<40; i++)
      {
         udelay(1000);
      }
   }
#endif

   /* First msg serial sent to DSP will be 1 (inc'd in mlb_send_to_dsp()) */
   msg_serial = 0;

   /* Check whether mailbox regs can now be accessed */
   coreCmIdleSt = reg32_read(CORE_CM_IDLEST1);

#if 0
   if(c64_debug_level >= 20)
   {
      printk(KERN_INFO "c64::mlb_init: CM_IDLEST1_CORE=0x%08x\n",
             coreCmIdleSt
             );
   }
#endif

   /* check ST_MAILBOXES bit */
   if(0 == (coreCmIdleSt & 0x80))
   {
      msg_base_virt =
         (sU8 __iomem *)
         ioremap_nocache(MSG_BASE,
                         DSP_RESET_VECTOR_MSGSIZE
                         );

      if(NULL != msg_base_virt)
      {
         mlb_base_virt = (sU8 __iomem *) ioremap_nocache(MLB_BASE, 4096);
         
         if(NULL != mlb_base_virt)
         {
#ifdef PANDORA_VID2_MSG_HACK
            hack_base_virt =
            (sU8 __iomem *)
               ioremap_nocache(DSP_HACK_BASE,
                               32
                               );
            
            if(NULL != hack_base_virt)
#endif /* PANDORA_VID2_MSG_HACK */
            {
               sU32 mlbRev;
               
               mlbRev = MLB_REG(MLB_REVISION);
#if 1
               if(c64_debug_level >= 10)
               {
                  printk(KERN_INFO "c64::mlb_init: mlb revision=0x%08x\n", mlbRev); 
               }
#endif
               
               /* Disable mailbox interrupts and reset mailbox */
               loc_mlb_reset();
               
               
               /* Install interrupt handler for GPP mailbox interrupt */
               {
                  int r;
                  
                  r = request_irq(MLB_U0_MPU_IRQ,
                                  &irq_handler__mlb,
                                  0,//IRQF_DISABLED,      // fast interrupt handler, (see interrupt.h:IRQF_xxx)
                                  "c64-mlb",          // dev-name (for /proc/interrupts)
                                  NULL                // dev-id (ptr. to priv. data)
                                  );
                  
                  if(0 == r)
                  {
                     ////enable_irq(MLB_U0_MPU_IRQ);
                     
                     /* Enable mailbox interrupts */
                     loc_mlb_reset_2();
                     
                     
#if 0
                     printk(KERN_INFO "c64::mlb_init: unmasking IVA2_IRQ_10.\n");
                     /* Unmask IVA2_IRQ_10 (mailbox interrupt) */
                     ////reg32_write(0x01c21070u, (1u << 10)); // --> bus error
                     printk(KERN_INFO "c64::mlb_init:   -> done.\n");
#endif
                     
                     atomic_set(&at_b_ready_to_send, S_TRUE);
                     
                     /* Succeeded */
                     ret = S_TRUE;
                  }
                  else
                  {
                     /* Failed: request_irq() */
                     printk(KERN_ERR "c64::mlb_init: request_irq() failed.\n");
                     
#ifdef PANDORA_VID2_MSG_HACK
                     iounmap(hack_base_virt);
                     hack_base_virt = NULL;
#endif /* PANDORA_VID2_MSG_HACK */

                     iounmap(mlb_base_virt);
                     mlb_base_virt = NULL;
                    
                     iounmap(msg_base_virt);
                     msg_base_virt = NULL;
                     
                     ret = S_FALSE;
                  }
               }
            }
#ifdef PANDORA_VID2_MSG_HACK
            else
            {
               /* Failed to map VID2 scaler registers */
               printk(KERN_ERR "c64::mlb_init: ioremap_nocache() [hack] (VID2 scaler) failed.\n");
               
               iounmap(mlb_base_virt);
               mlb_base_virt = NULL;
               
               iounmap(msg_base_virt);
               msg_base_virt = NULL;
               
               ret = S_FALSE;
            }
#endif /* PANDORA_VID2_MSG_HACK */
         }
         else
         {
            /* Failed: ioremap_nocache() [mlb] */
            printk(KERN_ERR "c64::mlb_init: ioremap_nocache() [mlb] failed.\n");
            
            iounmap(msg_base_virt);
            msg_base_virt = NULL;
            
            ret = S_FALSE;
         }
      }
      else
      {
         /* Failed: ioremap_nocache() [msg] */
         printk(KERN_ERR "c64::mlb_init: ioremap_nocache() [msg] failed.\n");
         
         ret = S_FALSE;
      }
   }
   else
   {
      /* Failed to enable access to mailbox registers */
      printk(KERN_ERR "c64::mlb_init: mailbox regs cannot be accessed.\n");

      ret = S_FALSE;
   }

   return ret;
}


/*--------------------------------------------------------------------------- mlb_exit() */
void mlb_exit(void) {

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::mlb_exit: ENTER  (%u irqs handled so far)\n", mlb_irq_count);
   }

   atomic_set(&at_b_ready_to_send, S_FALSE);

   ////disable_irq(MLB_U0_MPU_IRQ);

   /* Disable GPP mailbox interrupts */
   MLB_REG(MLB_IRQENABLE_0) = 0;

   /* Disable DSP mailbox interrupts */
   MLB_REG(MLB_IRQENABLE_1) = 0;

   /* Uninstall interrupt handler */
   free_irq(MLB_U0_MPU_IRQ,
            NULL             // dev-id (ptr. to priv. data)
            );

   /* Unmap mailbox registers */
   iounmap(mlb_base_virt);
   mlb_base_virt = NULL;

#ifdef PANDORA_VID2_MSG_HACK
   /* Unmap hack (VID2 scaler) registers */
   iounmap(hack_base_virt);
   hack_base_virt = NULL;
#endif /* PANDORA_VID2_MSG_HACK */

   /* Unmap DSP<>GPP shared memory area */
   iounmap(msg_base_virt);
   msg_base_virt = NULL;

   if(c64_debug_level >= 5)
   {
      printk(KERN_INFO "c64::mlb_exit: LEAVE\n");
   }
}
