/* ----
 * ---- file   : dsp.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the LESSER GNU GENERAL PUBLIC LICENSE (LGPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING_LGPL for further information.
 * ----
 * ---- info   : Arch-independent DSP utility functions. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013
 * ----
 * ----
 */

#include <string.h>

#include "../include/types.h"
#include "../include/log.h"

#include "../include/dsp_err.h"
#include "../include/dsp_common.h"
#include "../include/dsp_config.h"
#include "../include/dsp.h"
#include "../include/osal_err.h"
#include "../include/osal.h"


#ifndef HAVE_OS_LINUX
#error dsp.c must be compiled for Linux (no cygwin support here)
#endif


#define USE_FAST_MMAP defined




#ifdef USE_FAST_MMAP
static sBool b_dsp_mmap_force_real;


static dsp_mem_region_t fast_mmap[] = {

   // DSP memory (updated in dsp_init() with actual dsp_config)
   //   must be the first entry in this array!
   { 0,0,0 },

   // SCM / PRM DSP regs
   {
      .phys_addr = 0x48000000u,
      .size      = 0x00400000u
   }
};

#define NUM_FAST_MMAP_ENTRIES  (sizeof(fast_mmap) / sizeof(dsp_mem_region_t))

#endif // USE_FAST_MMAP


/*--------------------------------------------------------------------------- osal_strerror() */
const char * S_CALL dsp_strerror(dsp_error_t _err) {
   switch(_err)
   {
      default:
         return NULL;

      case DSP_ERR_OK:           return "DSP_ERR_OK";

      case DSP_ERR_MEM:          return "DSP_ERR_MEM";
      case DSP_ERR_MEM_OPEN:     return "DSP_ERR_MEM_OPEN";
      case DSP_ERR_MEM_MAP:      return "DSP_ERR_MEM_MAP";
      case DSP_ERR_MEM_UNMAP:    return "DSP_ERR_MEM_UNMAP";
      case DSP_ERR_MEM_CHECK:    return "DSP_ERR_MEM_CHECK";
   }
}


/*--------------------------------------------------------------------------- dsp_init() */
dsp_error_t dsp_init(void) {
   dsp_error_t ret;

   /* Initialize OS abstraction layer */
   ret = osal_init();

   if(DSP_ERR_OK == ret)
   {

#ifdef USE_FAST_MMAP
      // Init fast mmap
      {
         sUI mapIdx;

         b_dsp_mmap_force_real = S_TRUE;

         for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
         {
            fast_mmap[mapIdx].virt_addr = 0;
         }

         fast_mmap[0].phys_addr = dsp_config.reset_vector.phys_addr;
         fast_mmap[0].size      = (dsp_config.pool.phys_addr + dsp_config.pool.size) - dsp_config.reset_vector.phys_addr;

         for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
         {
            ret = dsp_mmap(fast_mmap[mapIdx].phys_addr, fast_mmap[mapIdx].size, &fast_mmap[mapIdx].virt_addr);

            if(DSP_ERR_OK == ret)
            {
               if(log_debug_level >= 10)
               {
                  log_printf(LOG_INFO "dsp_init: initialized fast_mmap[%u] physAddr=0x%08x size=0x%08x\n",
                             mapIdx,
                             fast_mmap[mapIdx].phys_addr,
                             fast_mmap[mapIdx].size
                             );
               }
            }
            else
            {
               log_printf(LOG_ERROR "dsp_init: failed to mmap() fast_mmap region phys_addr=0x%08x size=0x%08x\n",
                          fast_mmap[mapIdx].phys_addr,
                          fast_mmap[mapIdx].size
                          );
               break;
            }
            // Next entry
         }

         b_dsp_mmap_force_real = S_FALSE;
      }
#endif // USE_FAST_MMAP
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_exit() */
void dsp_exit(void) {

#ifdef USE_FAST_MMAP
   // Teardown fast mmap
   {
      sUI mapIdx;
      dsp_error_t err;
      b_dsp_mmap_force_real = S_TRUE;
      
      for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
      {
         if(0 != fast_mmap[mapIdx].virt_addr)
         {
            err = dsp_munmap(fast_mmap[mapIdx].virt_addr, fast_mmap[mapIdx].size);
            
            if(DSP_ERR_OK == err)
            {
               if(log_debug_level >= 10)
               {
                  log_printf(LOG_INFO "dsp_exit: de-initialized fast_mmap[%u] physAddr=0x%08x virtAddr=0x%08x size=0x%08x\n",
                             mapIdx,
                             fast_mmap[mapIdx].phys_addr,
                             fast_mmap[mapIdx].virt_addr,
                             fast_mmap[mapIdx].size
                             );
               }

            }
            else
            {
               log_printf(LOG_ERROR "dsp_exit: failed to munmap() fast_mmap region virt_addr=0x%08x size=0x%08x\n",
                          fast_mmap[mapIdx].virt_addr,
                          fast_mmap[mapIdx].size
                          );
               break;
            }
         }

         // Next entry
      }
   }
#endif // USE_FAST_MMAP

   /* Shutdown OS abstraction layer */
   osal_exit();
}


/*--------------------------------------------------------------------------- dsp_mmap() */
dsp_error_t dsp_mmap(sUI _physAddr, sUI _numBytes, sUI *_retAddr) {
   dsp_error_t ret;

   // printf("xxx dsp_mmap 1\n");

#ifdef USE_FAST_MMAP
   if(!b_dsp_mmap_force_real)
   {
      sUI mapIdx;

      for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
      {
         if( (_physAddr >= fast_mmap[mapIdx].phys_addr) &&
             (_physAddr <  (fast_mmap[mapIdx].phys_addr + fast_mmap[mapIdx].size))
             )
         {
            *_retAddr = (_physAddr - fast_mmap[mapIdx].phys_addr) + fast_mmap[mapIdx].virt_addr;

            if(log_debug_level >= 20)
            {
               log_printf(LOG_TRACE "**dsp_mmap(physAddr=0x%08x numBytes=0x%08x (%u))\n",
                          _physAddr,
                          _numBytes,
                          _numBytes
                          );
            }

            return DSP_ERR_OK;
         }
      }

      log_printf(LOG_WARN "dsp_mmap: fallback to mmap() physAddr=0x%08x numBytes=0x%08x\n",
                 _physAddr,
                 _numBytes
                 );
   }
#endif // USE_FAST_MMAP

   // printf("xxx dsp_mmap 2\n");

   if(log_debug_level >= 20)
   {
      // printf("xxx dsp_mmap 2a\n");
      log_printf(LOG_TRACE "dsp_mmap(physAddr=0x%08x numBytes=0x%08x (%u))\n",
                 _physAddr,
                 _numBytes,
                 _numBytes
                 );
   }

   // printf("xxx dsp_mmap 3\n");

   if(NULL != _retAddr)
   {
      
      union {
         void *ptr;
         sUI   u32;
      } virtAddr;

      sU32 physAddrAlign = DSP_PAGESIZE_ALIGN(_physAddr);
      sU32 numBytesRound = DSP_PAGESIZE_ROUND(_numBytes);

      // printf("xxx dsp_mmap 4\n");

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_mmap: physAddrAlign=0x%08x numBytesRound=0x%08x (%u)\n",
                    physAddrAlign,
                    numBytesRound,
                    numBytesRound
                    );
      }
 

      // printf("xxx dsp_mmap 5\n");

      virtAddr.ptr = osal_mmap((void*)physAddrAlign, (sSize)numBytesRound);

      // printf("xxx dsp_mmap 6 log_debug_level=%u\n", log_debug_level);

      if(NULL != virtAddr.ptr)
      {
         // Succeeded
         virtAddr.u32 = (_physAddr - physAddrAlign) + virtAddr.u32;

         ret = DSP_ERR_OK;
      }
      else
      {
         // Failed: mmap()
         log_puts(LOG_ERROR "dsp_mmap: osal_mmap() failed.\n");

         ret = DSP_ERR_MEM_MAP;
      }

      *_retAddr = virtAddr.u32;

      if(log_debug_level >= 10)
      {
         log_printf("xxx dsp_mmap() virtAddr=0x%08x\n", virtAddr.u32);
      }
   }
   else
   {
      // Failed: retAddr is NULL
      ret = DSP_ERR_MEM;
   }

   if(log_debug_level >= 20)
   {
      log_printf(LOG_TRACE "dsp_mmap: ret=%d\n", ret);
   }
   
   return ret;
}


/*--------------------------------------------------------------------------- dsp_munmap() */
dsp_error_t dsp_munmap(sUI _virtAddr, sUI _numBytes) {
   dsp_error_t ret;

#ifdef USE_FAST_MMAP
   if(!b_dsp_mmap_force_real)
   {
      sUI mapIdx;

      for(mapIdx=0; mapIdx < NUM_FAST_MMAP_ENTRIES; mapIdx++)
      {
         if( (_virtAddr >= fast_mmap[mapIdx].virt_addr) &&
             (_virtAddr <  (fast_mmap[mapIdx].virt_addr + fast_mmap[mapIdx].size))
             )
         {
            if(log_debug_level >= 20)
            {
               log_printf(LOG_TRACE "**dsp_munmap: virtAddr=0x%08x numBytes=0x%08x (%u)\n",
                          _virtAddr,
                          _numBytes,
                          _numBytes
                          );
            }

            return DSP_ERR_OK;
         }
      }

      log_printf(LOG_WARN "dsp_munmap: fallback to munmap()\n");
   }
#endif // USE_FAST_MMAP

   if(log_debug_level >= 20)
   {
      log_printf(LOG_TRACE "dsp_munmap: virtAddr=0x%08x numBytes=0x%08x (%u)\n",
                 _virtAddr,
                 _numBytes,
                 _numBytes
                 );
   }

   {
      sUI virtAddrAlign = DSP_PAGESIZE_ALIGN(_virtAddr);
      sUI numBytesRound = DSP_PAGESIZE_ROUND(_numBytes);
      
      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_munmap: virtAddrAlign=0x%08x numBytesRound=0x%08x (%u)\n",
                    virtAddrAlign,
                    numBytesRound,
                    numBytesRound
                    );
      }

      ret = osal_munmap((void*)virtAddrAlign, numBytesRound);
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_reg_write_32() */
dsp_error_t dsp_reg_write_32(sUI _physAddr, sUI _value) {
   dsp_error_t ret;
   sUI virtAddr;

   ret = dsp_mmap(_physAddr, 4, &virtAddr);

   if(DSP_ERR_OK == ret)
   {
      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_write_32: physAddr=0x%08x oldVal=0x%08x newVal=0x%08x\n",
                    _physAddr,
                    *((sUI*)virtAddr),
                    _value
                    );
      }

      *((sUI*)virtAddr) = _value;

      // Succeeded (well, almost)
      ret = dsp_munmap(virtAddr, 4);
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_write_32: dsp_mmap() failed (ret=0x%08x).\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_reg_read_32() */
dsp_error_t dsp_reg_read_32(sUI _physAddr, sUI *_retValue) {
   dsp_error_t ret;
   sUI virtAddr;

   if(NULL != _retValue)
   {
      ret = dsp_mmap(_physAddr, 4, &virtAddr);
      
      if(DSP_ERR_OK == ret)
      {
         *_retValue = *((sUI*)virtAddr);

         if(log_debug_level >= 20)
         {
            log_printf(LOG_TRACE "dsp_reg_read_32: physAddr=0x%08x retVal=0x%08x\n",
                       _physAddr, *((sUI*)virtAddr)
                       );
         }
         
         // Succeeded (well, almost)
         ret = dsp_munmap(virtAddr, 4);
      }
      else
      {
         log_printf(LOG_ERROR "dsp_reg_read_32: dsp_mmap() failed (ret=0x%08x).\n", ret);
      }
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_read_32: _retValue is NULL!!\n");
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_reg_bit_set() */
dsp_error_t dsp_reg_bit_set(sUI _physAddr, sUI _bitNr) {
   dsp_error_t ret;
   sUI virtAddr;

   ret = dsp_mmap(_physAddr, 4, &virtAddr);

   if(DSP_ERR_OK == ret)
   {
      sUI val = *((sUI*)virtAddr);

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_bit_set: physAddr=0x%08x bitNr=%u oldVal=0x%08x\n",
                    _physAddr, _bitNr, val
                    );
      }

      val |= (1 << _bitNr);

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_bit_set: physAddr=0x%08x bitNr=%u newVal=0x%08x\n",
                    _physAddr, _bitNr, val
                    );
      }

      *((sUI*)virtAddr) = val;

      // Succeeded (well, almost)
      ret = dsp_munmap(virtAddr, 4);
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_bit_set: dsp_mmap() failed (ret=0x%08x).\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_reg_bit_clear() */
dsp_error_t dsp_reg_bit_clear(sUI _physAddr, sUI _bitNr) {
   dsp_error_t ret;
   sUI virtAddr;

   ret = dsp_mmap(_physAddr, 4, &virtAddr);

   if(DSP_ERR_OK == ret)
   {
      sUI val = *((sUI*)virtAddr);

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_bit_clear: physAddr=0x%08x bitNr=%u oldVal=0x%08x\n",
                    _physAddr, _bitNr, val
                    );
      }

      val &= ~(1 << _bitNr);

      *((sUI*)virtAddr) = val;

      if(log_debug_level >= 20)
      {
         log_printf(LOG_TRACE "dsp_reg_bit_clear: physAddr=0x%08x bitNr=%u newVal=0x%08x\n",
                    _physAddr, _bitNr, val
                    );
      }

      // Succeeded (well, almost)
      ret = dsp_munmap(virtAddr, 4);
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_bit_clear: dsp_mmap() failed (ret=0x%08x).\n", ret);
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_copy_to() */
dsp_error_t dsp_copy_to(sUI _physAddr, void *_srcBuf, sUI _numBytes) {
   dsp_error_t ret;
   sUI virtAddr;

   if(NULL != _srcBuf)
   {
      ret = dsp_mmap(_physAddr, _numBytes, &virtAddr);
      
      if(DSP_ERR_OK == ret)
      {
         memcpy((void*)virtAddr, _srcBuf, (size_t)_numBytes);

#if 0
         sUI i;
         sU8 *d = (sU8*)virtAddr;
         sU8 *s = (sU8*)_srcBuf;

         for(i=0; i<_numBytes; i++)
         {
            sU8 sv = s[i];
            d[i] = sv;
         }
#endif
         
         // Succeeded (well, almost)
         ret = dsp_munmap(virtAddr, _numBytes);
      }
      else
      {
         log_printf(LOG_ERROR "dsp_reg_read_32: dsp_mmap() failed (ret=0x%08x).\n", ret);
      }
   }
   else
   {
      log_printf(LOG_ERROR "dsp_reg_read_32: _retValue is NULL!!\n");
   }

   return ret;
}


/*--------------------------------------------------------------------------- dsp_memset() */
dsp_error_t dsp_memset(sUI _physAddr, sU8 _c, sUI _numBytes) {
   dsp_error_t ret;
   sUI virtAddr;

   ret = dsp_mmap(_physAddr, _numBytes, &virtAddr);
 
   if(DSP_ERR_OK == ret)
   {
      memset((void*)virtAddr, _c, (size_t)_numBytes);
      
      // Succeeded (well, almost)
      ret = dsp_munmap(virtAddr, _numBytes);
   }
   else
   {
      log_printf(LOG_ERROR "dsp_memset: dsp_mmap() failed (ret=0x%08x).\n", ret);
   }
   
   return ret;
}
