/* ----
 * ---- file   : test_hello.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : This is part of the "c64_load" utility.
 * ----
 * ---- changed: 12Sep2013, 15Sep2013
 * ----
 * ----
 */

#define MAX_RPC_READ_RETRIES (1000u)

#define NUM_BENCHMARK_ITERATIONS (1000000u)
//#define NUM_BENCHMARK_ITERATIONS (10u)


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <fcntl.h>
#include <unistd.h>

#include <sys/time.h>

#include "../include/types.h"
#include "../include/log.h"
//#include "../include/osal_err.h"
//#include "../include/osal.h"
#include "../include/dsp_err.h"
//#include "../include/dsp_common.h"
//#include "../include/dsp_config.h"
#include "../include/dsp.h"
#include "../include/msg.h"

#include <c64p_simple_dm3730/pjt/components.h>


/*--------------------------------------------------------------------------- module vars */
static int fd;

static struct timeval tv_start;

static sUI rpc_count = 0;


/*--------------------------------------------------------------------------- loc_print_dspdebug() */
static void loc_print_dspdebug(void) {
   sU32 val;

   ssize_t numRead = read(fd, &val, 4);
   
   log_printf(LOG_DEBUG "DSP numRead=%u, debug=0x%08x\n", (sUI)numRead, val);
}


/*--------------------------------------------------------------------------- loc_milliseconds_init() */
void loc_milliseconds_init(void) {
   gettimeofday(&tv_start, 0);

   srand(tv_start.tv_usec);
}


/*--------------------------------------------------------------------------- loc_milliseconds() */
sUI loc_milliseconds_get(void) {
   struct timeval c;

   gettimeofday(&c, 0);

   return (sUI) ( (c.tv_sec - tv_start.tv_sec) * 1000 + (c.tv_usec - tv_start.tv_usec)/1000 );
}


/*--------------------------------------------------------------------------- loc_send_forced_message() */
static void loc_send_forced_message(sU32 _val) {
   
   ssize_t numWritten = write(fd, &_val, 4);

   //if(log_debug_level > 0)
   {
      log_printf(LOG_INFO "write testmsg: %u bytes written.\n", numWritten);
   }
   
   loc_print_dspdebug();
}


/*--------------------------------------------------------------------------- loc_rpc() */
static int loc_rpc(const c64_msg_t *_msgTo,
                        c64_msg_t *_msgFrom
                   ) {

   /* Send RPC message to DSP */
   int ret;
   ssize_t numWritten;

   numWritten = write(fd, _msgTo, sizeof(c64_msg_t));
   
   if(sizeof(c64_msg_t) == numWritten)
   {
      /* Now read result */
      ssize_t numRead;
      sUI retryCount = 0;
      
      do
      {
         /* (note) read() can fail with EAGAIN while DSP reply is not available, yet */
         numRead = read(fd, _msgFrom, sizeof(c64_msg_t));

         if(log_debug_level >= 10)
         {
            log_printf("loc_rpc[%08x]: [%04u] numRead=%u\n", rpc_count, retryCount, numRead);
         }

         retryCount++;
      }
      while((sizeof(c64_msg_t) != numRead) && (EAGAIN == errno) && (0 != numRead) && (retryCount++ < MAX_RPC_READ_RETRIES));

      if(MAX_RPC_READ_RETRIES == retryCount)
      {
         log_printf(LOG_ERROR "loc_rpc[%08x]: read() timeout after %u retries.\n",
                    rpc_count,
                    MAX_RPC_READ_RETRIES
                    );
      }
      
      if(sizeof(c64_msg_t) == numRead)
      {
         if(log_debug_level >= 10)
         {
            log_printf(LOG_INFO "loc_rpc[%08x]: ok, got reply from DSP. ret1=0x%08x ret2=0x%08x\n", 
                       rpc_count,
                       _msgFrom->data[0].ret,
                       _msgFrom->data[1].ret
                       );
         }
         
         /* Succeeded */
         ret = 0;
      }
      else
      {
         /* Failed: read() reply */
         log_printf(LOG_ERROR "loc_rpc[%08x]: read() failed. retryCount=%u numRead=%u errno=%d (\"%s\").\n", 
                    rpc_count,
                    retryCount,
                    numRead,
                    errno,
                    strerror(errno)
                    );
         
         loc_print_dspdebug();
         
         ret = 10;
      }
   }
   else
   {
      /* Failed: write() */
      log_printf(LOG_ERROR "loc_rpc[%08x]: write() failed. errno=%d (\"%s\").\n", 
                 rpc_count,
                 errno,
                 strerror(errno)
                 );
      
      ret = 10;
   }

   rpc_count++;

   return ret;
}


/*--------------------------------------------------------------------------- loc_calibrate() */
static void loc_calibrate(void) {
   sUI i;

   log_printf(LOG_INFO "starting calibration\n");

#if 0
   /* (note) sometimes the DSP triggers a message IRQ after being reset.
    *         Just send a couple of dummy messages by running this program w/o args
    *         to remedy this situation.
    *
    *         sending a forced message will reset the "current_client" field in
    *          the kernel module
    */
#endif

   ////usleep(1000);
   
   /* (note) for some reason, the first RPC always returns the wrong result (0) although the
    *         caches have been written back on the DSP-side and the shared message memory
    *         is uncached (at least it should be, ioremap_nocache() was used to map it in the kernel)
    */
   for(i=0; i<2; i++)
   {
      c64_msg_t msg;
      c64_msg_t reply;

      loc_rpc(&msg, &reply);

      //usleep(1000);
   }
}


/*--------------------------------------------------------------------------- loc_benchmark() */
static void loc_benchmark(void) {

   log_printf(LOG_INFO "starting benchmark\n");
   {
      c64_msg_t msg;
      c64_msg_t reply;
      sUI       tStart = loc_milliseconds_get();
      sUI       tEnd;
      sUI       tDelta;
      sUI       i;
      int       r;

      msg.req.dsp_component_id = COMPONENT_DEMO_CALC; // selects "demo_calc" component on DSP-side
      msg.req.dsp_cmd_id       = DEMO_CALC_CMD_ADD;   // selects the component's 'ADD' command
      
      msg.data[0].arg = 42;                           // argument #1
      msg.data[1].arg = 0;                            // argument #2

      for(i=0; i<NUM_BENCHMARK_ITERATIONS; i++)
      {
         r = loc_rpc(&msg, &reply);

         if(reply.data[0].ret != (msg.data[0].arg + msg.data[1].arg))
         {
            log_printf(LOG_ERROR "iteration=%u: unexpected RPC result: have=0x%08x, expected=0x%08x\n",
                       i,
                       reply.data[0].ret,
                       (msg.data[0].arg + msg.data[1].arg)
                       );

            loc_print_dspdebug();
            break;
         }

         /* Next iteration */
         msg.data[0].arg++;
         msg.data[1].arg++;
         ////usleep(2000);
      }

      if(NUM_BENCHMARK_ITERATIONS == i)
      {

         tEnd = loc_milliseconds_get();
         
         tDelta = tEnd - tStart;
         
         log_printf(LOG_INFO "benchmark finished. => %u iterations in %u millisecs.\n",
                    NUM_BENCHMARK_ITERATIONS,
                    tDelta
                    );
      }
      else
      {
         log_printf(LOG_ERROR "(benchmark aborted due to error)\n");
      }
   }
}


/*--------------------------------------------------------------------------- main() */
int main(int argc, char**argv) {
   int ret;
   
   //log_debug_level = 10;
   log_debug_level = 0;

   loc_milliseconds_init();

   fd = open("/dev/c64", O_RDWR | O_SYNC);

   if(-1 != fd)
   {
      loc_print_dspdebug();

      /* Send test message to DSP */
      if(1 == argc)
      {
         /* Send a message bypassing the client handler in the c64 kernel module 
          *
          *  This is just to see whether the GPP<>DSP messaging in general works OK. 
          *
          *   (note) the kernel module should have been loaded with "debug=20", i.e.
          *           $ /sbin/insmod c64.ko debug=20
          */
         loc_send_forced_message(0x37703370u);

         /* Succeeded */
         ret = 0;
      }
      else if(2 == argc)
      {
         /* Test single DSP remote procedure call */
         c64_msg_t msg;
         c64_msg_t reply;

         if(log_debug_level > 0)
         {
            log_printf(LOG_DEBUG "sizeof(c64_msg_t) == %u\n", sizeof(c64_msg_t)); // must be 12 bytes
         }

         msg.req.dsp_component_id = COMPONENT_DEMO_CALC; // selects "demo_calc" component on DSP-side
         msg.req.dsp_cmd_id       = DEMO_CALC_CMD_ADD;   // selects the component's 'ADD' command

         msg.data[0].arg = 42 + (rand()&31);             // argument #1
         msg.data[1].arg = 22 + (rand()&31);             // argument #2

         ret = loc_rpc(&msg, &reply);

         loc_print_dspdebug();

         if(0 == ret)
         {
            log_printf(LOG_INFO "log_rpc() OK: result=%d (expected %d)\n", 
                       *(sS32*) &reply.data[0].ret,   /* re-interpet unsigned int as signed */
                       (msg.data[0].arg + msg.data[1].arg)
                       );

            loc_print_dspdebug();
         }
         else
         {
            log_printf(LOG_ERROR "loc_rpc() failed (ret=%d)\n", ret);
         }
      }
      else
      {
         /* Benchmark DSP remote procedure call speed */

         loc_calibrate();

         loc_benchmark();
      }

      close(fd);
   }
   else
   {
      log_printf(LOG_ERROR "failed to open \"/dev/c64\".\n");
      ret = 10;
   }

   return ret;
}
